"""
The :mod:`pyrcn.elm` module includes Extreme Learning Machine algorithms.
"""

# See https://github.com/TUD-STKS/PyRCN for complete
# documentation.

# Author: Michael Schindler <michael.schindler1@mailbox.tu-dresden.de> with help from
#         the pyrcn community. ELM are copyright of their respective owners.
# License: BSD 3-Clause (C) TU Dresden 2020

from pyrcn.extreme_learning_machine._extreme_learning_machine import ELMClassifier, ELMRegressor

__all__ = ['ELMClassifier',
           'ELMRegressor'
           ]



#!/usr/bin/env python
# coding: utf-8

# # Onset detection using Echo State Networks
# 
# ## Introduction
# 
# In this notebook, we demonstrate how the ESN can deal with musical onset detection, detecting the start time of new note events.
# 
# As this is a computationally expensive task, we have pre-trained models to serve as an entry point.
# 
# At first, we import all packages required for this task. You can find the import statements below.

# In[1]:


import numpy as np
import scipy
import os
import csv
from collections import deque
from intervaltree import IntervalTree
from sklearn.base import clone
from joblib import dump, load

from matplotlib import pyplot as plt
plt.rcParams['image.cmap'] = 'jet'
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42
get_ipython().run_line_magic('matplotlib', 'inline')

from IPython.display import set_matplotlib_formats
set_matplotlib_formats('png', 'pdf')

import soundfile as sf
from madmom.utils import quantize_events
from madmom.processors import SequentialProcessor, ParallelProcessor
from madmom.audio import SignalProcessor, FramedSignalProcessor
from madmom.audio.stft import ShortTimeFourierTransformProcessor
from madmom.audio.filters import LogarithmicFilterbank
from madmom.audio.spectrogram import FilteredSpectrogramProcessor, LogarithmicSpectrogramProcessor, SpectrogramDifferenceProcessor

from pyrcn.echo_state_network import ESNRegressor


# ## Dataset
# 
# For training and evaluation, we have used the Böck dataset. Therefore, you need to download three archive: The annotations can be downloaded from ([https://drive.google.com/file/d/0B-MfhNTg9c5eTGJ6X0FOd0xDSWs/view](https://drive.google.com/file/d/0B-MfhNTg9c5eTGJ6X0FOd0xDSWs/view)), the split files can be downloaded from ([https://drive.google.com/file/d/0B-MfhNTg9c5eUV9DRVluNjdTOGs/view](https://drive.google.com/file/d/0B-MfhNTg9c5eUV9DRVluNjdTOGs/view)) and the audio files from ([https://drive.google.com/file/d/0B-MfhNTg9c5eeDktWUpjeUF3VkU/view](https://drive.google.com/file/d/0B-MfhNTg9c5eeDktWUpjeUF3VkU/view)). 
# 
# In the following, we assume that you have downloaded and extracted the dataset successfully to the subfolder "dataset/onset_detection". We have prepared the directory structure and one example file "ff123_ATrain.flac" to give you and idea how the dataset should be organized for this notebook.
# 
# The dataset is split into folds for 8-fold cross validation. The subdirectory "dataset/onset_detection/onsets_splits" contains eight file lists. Each contains 39 to 41 file names. By default, we have placed one file name in the first file list. The other ones are empty and need to be filled after downloading the datsaet.
# 
# We use the following function to return a nested list eight file lists, one for each fold:

# In[2]:


def get_splits(onsets_splits_path: str=r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset\onsets_splits"):
    """
    onsets_splits_path: str: /full/path/to/onsets_splits/
    """
    onsets_splits_path = os.path.normpath(onsets_splits_path)
    folds = []
    for _, _, file_names in os.walk(onsets_splits_path):
        for file_name in sorted(file_names):
            with open(os.path.join(onsets_splits_path, file_name)) as f:
                content = f.readlines()
                content = [x.strip() for x in content]
                folds.append(content)
    return folds


# ## Prepare an ESN model
# 
# We have already trained ESN models with different reservoir sizes in the uni- and bidirectional mode. 
# 
# In case you would like to train new models, we provide the code here. Note that this will take a lot of time as the MusicNet is a large-scale dataset. Furthermore, you need to download the dataset at first, as we have just included one audio file for demonstration purposes. 

# In[3]:


try:
    esn1 = load(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\models\esn_5000_bi_delta.joblib")
    esn2 = load(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\models\second_layer_esn_5000_bi.joblib")
except:
    esn1 = ESNRegressor(k_in=10, input_scaling=0.4, spectral_radius=0.5, bias=0.8, leakage=1.0, reservoir_size=500, k_res=10,
                        reservoir_activation='tanh', bi_directional=False, solver='ridge', beta=1e-2, random_state=1)
    esn2 = ESNRegressor(k_in=10, input_scaling=0.4, spectral_radius=0.5, bias=6.5, leakage=1.0, reservoir_size=500, k_res=10,
                        reservoir_activation='tanh', bi_directional=False, solver='ridge', beta=1e-2, random_state=1, ext_bias=1)


# ## Feature extraction
# 
# The acoustic features extracted from the input signal are obtained by filtering short-term spectra (window length 4096 samples and hop size 10 ms) with a bank of triangular filters in the frequency domain with log-spaced frequencies. The frequency range was 30 Hz to 17 000 Hz and we used 12 filters per octave. We used logarithmic magnitudes and added 1 inside the logarithm to ensure a minimum value of 0 for a frame without energy. The first derivative between adjacent frames was added in order to enrich the features by temporal information. Binary labels indicating absent (value 0) or present (value 1) pitches for each frame are assigned to each frame. Note that this task is a multilabel classification. Each MIDI pitch is a separate class, and multiple or no classes can be active at a discrete frame index.
# 
# For a more detailed description, please have a look in our repository ([https://github.com/TUD-STKS/Automatic-Music-Transcription](https://github.com/TUD-STKS/Automatic-Music-Transcription)) with several detailed examples for music analysis tasks.

# In[4]:


def superflux_delta(S, max_size: int = 3, lag: int = 2):
    init = np.empty((lag, S.shape[1]))
    init[:] = np.inf
    S = np.insert(S, 0, init, axis=0)
    filt_size = [1, max_size]
    ref = scipy.ndimage.maximum_filter(S, size=filt_size)
    diff = np.maximum(S[lag:] - ref[:-lag], 0.0)
    return diff


def extract_features(in_folder, file_name):
    full_path_to_audio = os.path.join(in_folder, 'onsets_audio', file_name + '.flac')
    full_path_to_label = os.path.join(in_folder, 'onsets_annotations', file_name + '.onsets')
    y, sr = sf.read(file=full_path_to_audio)
    if y.ndim > 1:
        y = y[:, 0]
    y = y / np.max(np.abs(y))

    sig = SignalProcessor(num_channels=1, sample_rate=sr)
    multi = ParallelProcessor([])
    for frame_size in [1024, 2048, 4096]:
        frames = FramedSignalProcessor(frame_size=frame_size, fps=100)
        stft = ShortTimeFourierTransformProcessor()  # caching FFT window
        filt = FilteredSpectrogramProcessor(filterbank=LogarithmicFilterbank, num_bands=7, fmin=27.5, fmax=16000., 
                                            norm_filters=False, unique_filters=True)
        spec = LogarithmicSpectrogramProcessor(log=np.log10, add=1.0)
        # process each frame size with spec and diff sequentially
        multi.append(SequentialProcessor((frames, stft, filt, spec)))
    stack = list
    pre_processor = SequentialProcessor((sig, multi, stack))
    
    spectrograms = [np.asarray(spectrogram) for spectrogram in pre_processor.process(data=y)]
    S = np.hstack(spectrograms)
    deltas = [superflux_delta(spectrogram, max_size=3, lag=2) for spectrogram in spectrograms]
    S = np.append(S, np.hstack(deltas), axis=1)
    deltas_deltas = [superflux_delta(delta, max_size=3, lag=2) for delta in deltas]
    S = np.append(S, np.hstack(deltas_deltas), axis=1)
    
    S = S - 1.0
    onset_times = np.loadtxt(fname=full_path_to_label)
    labels = quantize_events(events=onset_times, fps=100, length=S.shape[0])
    idx = np.argwhere(labels)
    labels[idx[idx < idx.shape[0]] + 1] = 0.5
    labels[idx[idx > 0] - 1] = 0.5
    return S, labels, onset_times


# ## Train a model
# 
# This might take a long time, because we stay in line with the reference algorithms and use 8-fold cross validation. 
# 
# The pipeline is clear: 
# 
# - Extract features and labels
# - Pass features and labels through ESN
# - Compute output weights and clear no more required helper variables.
# 
# Notice that you can pass the option "update_output_weights=False" to the call to partial_fit. If this option is passed, no output weights are computed after passing the sequence through the network. This is computationally more efficient.
# 
# After passing the entire sequence through the ESN, we need to call finalize() in order to compute the update weights.
# 
# By default, "update_output_weights=False" and it is not necessary to call finalize().
# 
# Here, we show just one step of the cross validation for the sake of demonstration. In the paper, we trained eight models for the 8-fold cross validation. Each was tested on the last fold in the current fold rotation.

# In[5]:


should_train = False

if should_train:
    folds = get_splits()
    for curr_split in range(len(folds)):
        curr_folds = deque(folds)
        curr_folds.rotate(curr_split)
        for d in range(len(curr_folds) - 1):
            train_files = curr_folds[d]
            for fid in train_ids:
                X, y_true, _ = extract_features(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset", file_name=fid)
                esn1.partial_fit(X=X, y=y_true, update_output_weights=False)
        esn1.finalize()
        break  # Here, we have trained a model on seven folds. It can be now evaluated on the training and test examples.


# ## Validate the ESN model
# 
# This might take a long time, because we stay in line with the reference algorithms and use 8-fold cross validation. 
# 
# The pipeline is similar as before: 
# 
# - Extract features and labels
# - Pass features through ESN and compute output
# - Save labels, outputs, binarized outputs for further processing, e.g. visualization or evaluation.

# In[6]:


folds = get_splits()
X_train = []
Onset_times_train = []
Y_pred_train = []

X_test = []
Onset_times_test = []
Y_pred_test = []

for curr_split in range(len(folds)):
    curr_folds = deque(folds)
    curr_folds.rotate(curr_split)
    for d in range(len(curr_folds) - 1):
        train_files = curr_folds[d]
        for fid in train_files:
            X, y_true, onset_times = extract_features(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset", file_name=fid)
            Onset_times_train.append(onset_times)
            X_train.append(X)
            y_pred = esn1.predict(X=X)
            Y_pred_train.append(y_pred)
    
    test_files = curr_folds[-1]
    X, y_true, onset_times = extract_features(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset", file_name=fid)
    Onset_times_test.append(onset_times)
    X_test.append(X)
    y_pred = esn1.predict(X=X)
    Y_pred_test.append(y_pred)
    break


# ## Visualization
# 
# We visualize inputs, onset detection function (ODF) and target onsets in two plots, respectively. 

# In[7]:


fig, (ax0, ax1) = plt.subplots(2, 1, figsize=(12, 6))

ax0.imshow(X_train[0].T, origin='lower', vmin=0, vmax=1, aspect='auto')
ax0.set_xlabel('n')
ax0.set_ylabel('X[n]')
ax0.grid()
for target in Onset_times_train[0]:
    ax0.axvline(100*target, color='w')

if Y_pred_train[0].ndim > 1:
    ax1.plot(Y_pred_train[0][:, 0])
else:
    ax1.plot(Y_pred_train[0][:, 0])
ax1.set_xlabel('n')
ax1.set_ylabel('y_pred[n]')
ax1.grid()
for target in Onset_times_train[0]:
    ax1.axvline(100*target, color='k')

plt.tight_layout()


# ## Multilayer ESN
# 
# In our paper, we have presented a novel way of stacking Echo State Networks. The ESN can be trained layer-wise. That means, we can finalize the training of the first layer and then train the second layer using the first layer.
# 
# We have proposed to use a second reservoir in the following way:
# 
# - Pass features through the first ESN and compute output
# - Pass features through the second layer, and use the output of the first ESN as bias for the second ESN.
# - Compute output of the second layer.
# 
# As can be seen in the code, this is rather straight forward to do. In general, we allow the ESN to receive several external bias inputs, by ext_bias > 0. In that case, the ext_bias last features are used as the external bias and passed through the ESN via the bias weight matrix. 
# 
# To implement this for onset detection:
# 
# - We compute the output of esn1. 
# - The feature vectors and the ouput of esn1 are concatenated.
# - The concatenated feature matrix is fed through esn2 and the outputs are computed.

# In[8]:


should_train = False

if should_train:
    folds = get_splits()
    for curr_split in range(len(folds)):
        curr_folds = deque(folds)
        curr_folds.rotate(curr_split)
        for d in range(len(curr_folds) - 1):
            train_files = curr_folds[d]
            for fid in train_ids:
                X, y_true, _ = extract_features(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset", file_name=fid)
                esn1.partial_fit(X=X, y=y_true, update_output_weights=False)
        esn1.finalize()
    
        break  # Here, we have trained a model on seven folds. It can be now evaluated on the training and test examples.

    folds = get_splits()
    for curr_split in range(len(folds)):
        curr_folds = deque(folds)
        curr_folds.rotate(curr_split)
        for d in range(len(curr_folds) - 1):
            train_files = curr_folds[d]
            for fid in train_ids:
                X, y_true, _ = extract_features(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset", file_name=fid)
                y_pred = esn1.predict(X=X)
                if y_pred.ndim > 1:
                    X = np.concatenate((X, np.atleast_2d(y_pred[:, 0]).T), axis=1)
                else:
                    X = np.concatenate((X, np.atleast_2d(y_pred).T), axis=1)
                esn2.partial_fit(X=X, y=y_true, update_output_weights=False)
        esn2.finalize()
    
        break  # Here, we have trained a model on seven folds. It can be now evaluated on the training and test examples.


# In[9]:


folds = get_splits()
X_train = []
Onset_times_train = []
Y_pred_train = []

X_test = []
Onset_times_test = []
Y_pred_test = []

for curr_split in range(len(folds)):
    curr_folds = deque(folds)
    curr_folds.rotate(curr_split)
    for d in range(len(curr_folds) - 1):
        train_files = curr_folds[d]
        for fid in train_files:
            X, y_true, onset_times = extract_features(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset", file_name=fid)
            Onset_times_train.append(onset_times)
            X_train.append(X)
            y_pred = esn1.predict(X=X)
            if y_pred.ndim > 1:
                X = np.concatenate((X, np.atleast_2d(y_pred[:, 0]).T), axis=1)
            else:
                X = np.concatenate((X, np.atleast_2d(y_pred).T), axis=1)
            y_pred = esn2.predict(X=X)
            Y_pred_train.append(y_pred)
    
    test_files = curr_folds[-1]
    X, y_true, onset_times = extract_features(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\onset_detection\boeck_dataset", file_name=fid)
    Onset_times_test.append(onset_times)
    X_test.append(X)
    y_pred = esn1.predict(X=X)
    Y_pred_test.append(y_pred)
    break


# In[10]:


fig, (ax0, ax1) = plt.subplots(2, 1, figsize=(12, 6))

ax0.imshow(X_train[0].T, origin='lower', vmin=0, vmax=1, aspect='auto')
ax0.set_xlabel('n')
ax0.set_ylabel('X[n]')
ax0.grid()
for target in Onset_times_train[0]:
    ax0.axvline(100*target, color='w')

if Y_pred_train[0].ndim > 1:
    ax1.plot(Y_pred_train[0][:, 0])
else:
    ax1.plot(Y_pred_train[0][:, 0])
ax1.set_xlabel('n')
ax1.set_ylabel('y_pred[n]')
ax1.grid()
for target in Onset_times_train[0]:
    ax1.axvline(100*target, color='k')

plt.tight_layout()




#!/usr/bin/env python
# coding: utf-8

# # Timeseries prediction of the Mackey-Glass Equation with ESNs 
# 
# ## Introduction
# 
# The Mackey-Glass system is essentially the differential equation, where we set the parameters to $\alpha = 0.2$, $\beta = 10$, $\gamma = 0.1$ and the time delay $\tau = 17$ in  order to have a mildly chaotic attractor. 
# 
# \begin{align}
# \label{eq:MackeyGlass}
# \dot{y}(t) = \alpha y(t-\tau) / (1 + y(t - \tau)^{\beta}) - \gamma y(t)
# \end{align}

# In[1]:


import numpy as np
from sklearn.metrics import mean_squared_error

from matplotlib import pyplot as plt
plt.rcParams['image.cmap'] = 'jet'
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42

from IPython.display import set_matplotlib_formats
set_matplotlib_formats('png', 'pdf')

from pyrcn.echo_state_network import ESNRegressor


# Load the dataset

# In[2]:


data = np.loadtxt("./dataset/MackeyGlass_t17.txt")


# The first 500 samples are visualized.

# In[3]:


plt.figure()
plt.plot(data[:500])
plt.xlabel("n")
plt.ylabel("X[n]")
plt.grid()


# Standardization -> From here on, we have a numpy array!!!

# In[4]:


data = data / (data.max() - data.min())


# Define Train/Test lengths

# In[5]:


initLen = 100 # number of time steps during which internal activations are washed-out during training
# we consider trainLen including the warming-up period (i.e. internal activations that are washed-out when training)
trainLen = initLen + 1900 # number of time steps during which we train the network
testLen = 2000 # number of time steps during which we test/run the network


# Echo State Network preparation

# In[6]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=1.2, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=500, k_res=10, wash_out=0, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1., teacher_shift=0., solver='ridge', beta=1e-4, random_state=0)


# Training and Prediction. Be careful, this can take a longer time!!!
# 
# The lowest MSE obtained with this settings were \num{5.97e-06} for the training set and \num{43.1e-06} for the test set.

# In[7]:


train_in = data[None,0:trainLen]
train_out = data[None,0+1:trainLen+1]
test_in = data[None,trainLen:trainLen+testLen]
test_out = data[None,trainLen+1:trainLen+testLen+1]

train_in, train_out = train_in.T, train_out.T
test_in, test_out = test_in.T, test_out.T

esn.fit(X=train_in, y=train_out)
train_pred = esn.predict(X=train_in)
test_pred = esn.predict(X=test_in)

train_err = mean_squared_error(y_true=train_out, y_pred=train_pred)
test_err = mean_squared_error(y_true=test_out, y_pred=test_pred)

print("Train MSE:\t{0}".format(train_err))
print("Test MSE:\t{0}".format(test_err))


# Prediction of the training set.

# In[8]:


plt.figure()
plt.plot(train_out)
plt.plot(train_pred)
plt.xlabel("n")
plt.ylabel("X[n]")


# Prediction of the test set.
# 

# In[9]:


plt.figure()
plt.plot(test_out)
plt.plot(test_pred)
plt.xlabel("n")
plt.ylabel("X[n]")




"""
The :mod:`pyrcn.echo_state_network` module includes Echo State Network algorithms.
"""

# See https://github.com/TUD-STKS/PyRCN for complete
# documentation.

# Author: Peter Steiner <peter.steiner@tu-dresden.de> with help from
#         the pyrcn community. ESNs are copyright of their respective owners.
# License: BSD 3-Clause (C) TU Dresden 2020

from pyrcn.echo_state_network._echo_state_network import ESNClassifier, ESNRegressor

__all__ = ['ESNClassifier',
           'ESNRegressor'
           ]



import os
import glob
import numpy as np
from tqdm import tqdm
import time

from sklearn.base import clone
from sklearn.model_selection import train_test_split, ParameterGrid
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.metrics import mean_squared_error, classification_report, confusion_matrix, ConfusionMatrixDisplay, accuracy_score
from sklearn.cluster import MiniBatchKMeans
from joblib import Parallel, delayed
from pyrcn.echo_state_network import ESNRegressor
import matplotlib.pyplot as plt

import librosa


all_files = glob.glob(r"E:\free-spoken-digit-dataset\recordings\*.wav")
print(len(all_files))

X = [None] * len(all_files)
y = [None] * len(all_files)
print("extracting features...")
with tqdm(total=len(all_files)) as pbar:
    for k, f in enumerate(all_files):
        basename = os.path.basename(f).split('.')[0]
        # Get label (0-9) of recording.
        label = int(basename.split('_')[0])
        # Load the audio signal and normalize it.
        x, sr = librosa.core.load(f, sr=None, mono=False)
        # x /= np.max(np.abs(x))
        mfcc = librosa.feature.mfcc(y=x, sr=sr, hop_length=int(0.01*sr), n_fft=256, htk=True, n_mels=100, n_mfcc=13)
        mfcc_delta = librosa.feature.delta(mfcc)
        mfcc_delta2 = librosa.feature.delta(mfcc, order=2)
        X[k] = mfcc.T
        y[k] = label
        pbar.update(1)
print("done!")


X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=1, stratify=y)
X_val, X_test, y_val, y_test = train_test_split(X_test, y_test, test_size=0.5, random_state=1, stratify=y_test)

# Validate training and test sizes
print(len(X_train), len(y_train), X_train[0].shape, y_train[0])
print(len(X_val), len(y_val), X_val[0].shape, y_val[0])
print(len(X_test), len(y_test), X_test[0].shape, y_test[0])


# Normalize all features using the StandardScaler from scikit-learn.
scaler = StandardScaler().fit(X=np.vstack(X_train))
X_train_scaled = [scaler.transform(X) for X in X_train]
X_val_scaled = [scaler.transform(X) for X in X_val]
X_test_scaled = [scaler.transform(X) for X in X_test]


# One-Hot encoding of labels
enc = OneHotEncoder(sparse=False).fit(X=np.asarray(y_train).reshape(-1, 1))


# Clone the base_esn and fit it on the training data
def opt_function(base_reg, encoder, params, X_train, y_train, X_test, y_test, w_in=None):
    reg = clone(base_reg).set_params(**params)
    if w_in is not None:
        reg.initialize_from_outside(y=enc.transform(np.asarray(y_train[0]).reshape(-1, 1)),
                                    n_features=X_train[0].shape[0], input_weights=w_in, reservoir_weights=None,
                                    bias_weights=None)
    for X, y in zip(X_train, y_train):
        y = encoder.transform(np.asarray(y).reshape(1, -1))
        y = np.repeat(np.atleast_2d(y), repeats=X.shape[0], axis=0)
        reg.partial_fit(X=X, y=y, update_output_weights=False)
    reg.finalize()
    err_train = []
    Y_true_train = []
    Y_pred_train = []
    for X, y in zip(X_train, y_train):
        Y_true_train.append(y)
        y_true = encoder.transform(np.asarray(y).reshape(1, -1))
        y_true = np.repeat(np.atleast_2d(y_true), repeats=X.shape[0], axis=0)
        y_pred = reg.predict(X=X, keep_reservoir_state=False)
        err_train.append(mean_squared_error(y_true, y_pred))
        Y_pred_train.append(np.argmax(y_pred.sum(axis=0)))
    err_test = []
    Y_true_test = []
    Y_pred_test = []
    for X, y in zip(X_test, y_test):
        Y_true_test.append(y)
        y_true = encoder.transform(np.asarray(y).reshape(1, -1))
        y_true = np.repeat(np.atleast_2d(y_true), repeats=X.shape[0], axis=0)
        y_pred = reg.predict(X=X, keep_reservoir_state=False)
        err_test.append(mean_squared_error(y_true, y_pred))
        Y_pred_test.append(np.argmax(y_pred.sum(axis=0)))
    # return [np.mean(err_train), np.mean(err_test)]
    return [1 - accuracy_score(Y_true_train, Y_pred_train), 1 - accuracy_score(Y_true_test, Y_pred_test)]


"""
base_esn = ESNRegressor(k_in=-1, input_scaling=0.1, spectral_radius=0.0, bias=0.0, leakage=1.0, reservoir_size=1000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'input_scaling': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=-1, input_scaling=0.3, spectral_radius=0.0, bias=0.0, leakage=1.0, reservoir_size=1000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'leakage': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=-1, input_scaling=0.3, spectral_radius=0.0, bias=0.0, leakage=0.1, reservoir_size=1000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'input_scaling': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5],
        'spectral_radius': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=-1, input_scaling=0.8, spectral_radius=0.3, bias=0.0, leakage=0.1, reservoir_size=1000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'bias': [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=-1, input_scaling=0.8, spectral_radius=0.3, bias=0.6, leakage=0.1, reservoir_size=1000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'k_res': [5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=-1, input_scaling=0.8, spectral_radius=0.3, bias=0.6, leakage=0.1, reservoir_size=1000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

esn = clone(base_esn)
print("Train the ESN model...")
with tqdm(total=len(X_train_scaled)) as pbar:
    for X, y in zip(X_train_scaled, y_train):
        y = enc.transform(np.asarray(y).reshape(1, -1))
        y = np.repeat(np.atleast_2d(y), repeats=X.shape[0], axis=0)
        esn.partial_fit(X=X, y=y, update_output_weights=False)
        pbar.update(1)
esn.finalize()
print("... done!")

Y_true_train = []
Y_pred_train = []
print("Test the ESN model on the training data...")
with tqdm(total=len(X_train_scaled)) as pbar:
    for X, y in zip(X_train_scaled, y_train):
        Y_true_train.append(y)
        y_pred = esn.predict(X=X, keep_reservoir_state=False)
        Y_pred_train.append(np.argmax(y_pred.sum(axis=0)))
        pbar.update(1)
cm = confusion_matrix(Y_true_train, Y_pred_train)
cm_display = ConfusionMatrixDisplay(cm, display_labels=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).plot()
print("Classification training report for estimator %s:\n%s\n" % (esn, classification_report(Y_true_train, Y_pred_train)))
plt.show()

Y_true_test = []
Y_pred_test = []
print("Test the ESN model on the test data...")
with tqdm(total=len(X_test_scaled)) as pbar:
    for X, y in zip(X_test_scaled, y_test):
        Y_true_test.append(y)
        y_pred = esn.predict(X=X, keep_reservoir_state=False)
        Y_pred_test.append(np.argmax(y_pred.sum(axis=0)))
        pbar.update(1)
cm = confusion_matrix(Y_true_test, Y_pred_test)
cm_display = ConfusionMatrixDisplay(cm, display_labels=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).plot()
print("Classification test report for estimator %s:\n%s\n" % (esn, classification_report(Y_true_test, Y_pred_test)))
plt.show()
"""

kmeans = MiniBatchKMeans(n_clusters=500, reassignment_ratio=0, max_no_improvement=50, init='k-means++', verbose=1, random_state=0)
kmeans.fit(X=np.vstack(X_train_scaled))

new_input_weights = kmeans.cluster_centers_
# To compute the norm of the cluster centers, use the following line:
w_in = np.pad(np.divide(kmeans.cluster_centers_, np.linalg.norm(kmeans.cluster_centers_, axis=1)[:, None]), ((0, 3500), (0, 0)), mode='constant', constant_values=0)
# w_in = np.divide(kmeans.cluster_centers_, np.linalg.norm(kmeans.cluster_centers_, axis=1)[:, None])

base_esn = ESNRegressor(k_in=10, input_scaling=0.1, spectral_radius=0.0, bias=0.0, leakage=1.0, reservoir_size=4000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

"""
grid = {'input_scaling': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_val_scaled, y_val, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=10, input_scaling=0.4, spectral_radius=0.0, bias=0.0, leakage=1.0, reservoir_size=3000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'leakage': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))
"""
base_esn = ESNRegressor(k_in=10, input_scaling=0.4, spectral_radius=0.0, bias=0.0, leakage=0.1, reservoir_size=4000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'input_scaling': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5],
        'spectral_radius': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=10, input_scaling=1.2, spectral_radius=0.6, bias=0.0, leakage=0.1, reservoir_size=4000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'bias': [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=10, input_scaling=1.2, spectral_radius=0.6, bias=0.7, leakage=0.1, reservoir_size=4000,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

grid = {'k_res': [5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80]
       }

t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_esn, enc, params, X_train_scaled, y_train, X_test_scaled, y_test, w_in) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training DER: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest validation DER: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))

base_esn = ESNRegressor(k_in=10, input_scaling=1.2, spectral_radius=0.6, bias=0.7, leakage=0.1, reservoir_size=4000,
                        k_res=5, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=1)

esn = clone(base_esn)
esn.initialize_from_outside(y=enc.transform(np.asarray(y_train[0]).reshape(-1, 1)), n_features=X_train[0].shape[0],
                            input_weights=w_in, reservoir_weights=None, bias_weights=None)
print("Train the ESN model...")
with tqdm(total=len(X_train_scaled)) as pbar:
    for X, y in zip(X_train_scaled, y_train):
        y = enc.transform(np.asarray(y).reshape(1, -1))
        y = np.repeat(np.atleast_2d(y), repeats=X.shape[0], axis=0)
        esn.partial_fit(X=X, y=y, update_output_weights=False)
        pbar.update(1)
esn.finalize()
print("... done!")

Y_true_train = []
Y_pred_train = []
print("Test the ESN model on the training data...")
with tqdm(total=len(X_train_scaled)) as pbar:
    for X, y in zip(X_train_scaled, y_train):
        Y_true_train.append(y)
        y_pred = esn.predict(X=X, keep_reservoir_state=False)
        Y_pred_train.append(np.argmax(y_pred.sum(axis=0)))
        pbar.update(1)
cm = confusion_matrix(Y_true_train, Y_pred_train)
cm_display = ConfusionMatrixDisplay(cm, display_labels=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).plot()
print("Classification training report for estimator %s:\n%s\n" % (esn, classification_report(Y_true_train, Y_pred_train)))
plt.show()

Y_true_test = []
Y_pred_test = []
print("Test the ESN model on the test data...")
with tqdm(total=len(X_test_scaled)) as pbar:
    for X, y in zip(X_test_scaled, y_test):
        Y_true_test.append(y)
        y_pred = esn.predict(X=X, keep_reservoir_state=False)
        Y_pred_test.append(np.argmax(y_pred.sum(axis=0)))
        pbar.update(1)
cm = confusion_matrix(Y_true_test, Y_pred_test)
cm_display = ConfusionMatrixDisplay(cm, display_labels=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).plot()
print("Classification test report for estimator %s:\n%s\n" % (esn, classification_report(Y_true_test, Y_pred_test)))
plt.show()






#!/usr/bin/env python
# coding: utf-8

# # Stock Price Prediction using Echo State Networks
# 
# ## Introduction
# 
# In this notebook, the impact of different hyper-parameters of an ESN are explained. The notebook depends on just a small variety of packages: numpy, matplotlib, IPython and pyrcn.

# In[1]:


import numpy as np

from matplotlib import pyplot as plt
plt.rcParams['image.cmap'] = 'jet'
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42

from IPython.display import set_matplotlib_formats
set_matplotlib_formats('png', 'pdf')

from pyrcn.echo_state_network import ESNRegressor


# In this notebook, we feed an example impulse through the ESN, treating it as a non-linear filter. The unit impuls starts at n=5.

# In[2]:


X = np.zeros(shape=(100, 1), dtype=int)
X[5] = 1
y = X[:, 0]

plt.figure()
plt.plot(X)
plt.xlim([0, 100])
plt.ylim([0, 1])
plt.xlabel('n')
plt.ylabel('X[n]')
plt.grid()


# At first, show the impact of different input scaling factors.
# 
# Therefore, we neutralize the other hyper-parameters, i.e., no recurrent connections ($\rho = 0$), no bias ($\alpha_{\mathrm{b}} = 0$) and no leakage ($\lambda = 1$). 

# In[3]:


esn = ESNRegressor(k_in=1, input_scaling=0.1, spectral_radius=0.0, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# Visualizing this, we can see exactly what we might expect. We have chosen an input scaling factor of 0.1. Thus, the reservoir state is non-zero for exactly one sample. We can see that all reservoir states are zero all the times except for $n=5$, when the impulse is fed into the ESN. 
# 
# The absolute values of the reservoir states lie between 0 and 0.1.

# In[4]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# What happens if we increase the input scaling factor?

# In[5]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=0.0, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# Each reservoir state still has only one non-zero value at $n=5$ as before, just with higher activations up to 0.8. The $\tanh$ non-linearity is damping the reservoir states so that they cannot reach 1.  

# In[6]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# Let's keep the input scaling factor fixed to 1.0 for the next experiments. Echo State Networks have two ways to
# incorporate past information.
# 
# Next, let us analyze the spectral radius. Therefore, we set it to 0.3 in the following example.

# In[7]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=0.3, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# We can observe that the impulse responses are starting at n=5 and decaying until reaching zero after a short time. Obviously, the reservoir states are decaying rather fast, because the recurrent connections are small compared to the input scaling. 

# In[8]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# It is interesting to increase the spectral radius. Therefore, we set it to 0.9.

# In[9]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=0.9, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# The values are still bounded between [0.8, 0.8], but the reservoir states are active over a longer time now ($n\approx 30$).

# In[10]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# We can try to further increase the spectral radius. Therefore, we set it to 1.0.

# In[11]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=1.0, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# As we expected, the values are getting slightly larger, but are still bounded because of the tanh()-non-linearities.
# Interestingly, the reservoir states are not completely decaying, although the spectral radius was just increased by 0.1.
# 
# We can see that the reservoir states are decaying very slowly, and they are oscillating with a resonance frequency. For many tasks, it is indeed necessary to preserve the echo state property of reservoir and keep $\rho < 1$. However in some cases, such as time-series prediction in this paper, the spectral radius can be larger than 1. 

# In[12]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# For some use cases, the reservoir should behave more non-linear. Therefore, we can play around with the bias. Here, we
# will show the impact of the bias scaling factor just for one example and increase it to 0.2.
# 
# The spectral radius is decreased to 0.9 in order to fulfil the Echo State Property.

# In[13]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=0.9, bias=0.2, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# Two impacts of the bias scaling can be mainly observed: (1) The absolute value of the stable states of the reservoir neurons is approximately distributed from 0 to 0.2 and each neuron has its own stable state. When new information from the input is passed to the reservoir neurons, this is the excitation point. (2) Before the impulse arrives in the reservoir ($n=5$), the states are approaching their stable state. Due to the spectral radius, each reservoir neuron is connected to other neurons and thus feeds the constant bias through the network, until each neuron has reached its final state.

# In[14]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# Finally, let us observe the impact of the leakage. Leaky integration is the other way to incorporate past information
# into the reservoir. This works by keeping the previous reservoir states over a longer time.

# In[15]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=0.9, bias=0.0, ext_bias=False, leakage=0.3,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# The leakage behaves in the same way for all nodes in the reservoir and acts like a low-pass filter. The magnitude is strongly damped, and all reservoir states are decaying exponentially over a longer time. Due to the spectral radius, all neurons have individual decaying times.

# In[16]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# If we would like to incorporate future information into the reservoir, we can pass our input samples for- and backward
# through the reservoir. Therefore, we set bidirectional to True.

# In[17]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=0.9, bias=0.2, ext_bias=False, leakage=0.3,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=True,
                   teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# We can see the clear effect of usind bidirectional ESNs:
# 
# Because of the additional backward-pass, the number of reservoir states is doubled, and we can see that they decay in
# both, forward and backward direction. This is especially useful for some classification tasks.

# In[18]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:100, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# The ESN from the Mackey-Glass dataset with a reduced number of neurons

# In[19]:


esn = ESNRegressor(k_in=1, input_scaling=1.0, spectral_radius=1.2, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, wash_out=0, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1., teacher_shift=0., solver='ridge', beta=1e-4, random_state=0)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# Visualization of the Mackey-Glass-ESN

# In[20]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# The ESN from the Stock-Price dataset with a reduced number of neurons

# In[21]:


esn = ESNRegressor(k_in=1, input_scaling=0.6, spectral_radius=0.9, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1., teacher_shift=0., solver='ridge', beta=1e-8, random_state=0)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# Visualization of the Stock-Price-ESN

# In[22]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()


# The ESN from the Multipitch dataset with a reduced number of neurons

# In[23]:


esn = ESNRegressor(k_in=1, input_scaling=0.6, spectral_radius=0.2, bias=0.7, ext_bias=False, leakage=0.3,
                   reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1., teacher_shift=0., solver='ridge', beta=1e-8, random_state=0)

esn.fit(X=X, y=y, n_jobs=0)
_ = esn.predict(X=X, keep_reservoir_state=True)


# Visualization of the Multipitch-ESN

# In[24]:


plt.figure()
im = plt.imshow(np.abs(esn.reservoir_state[:50, 1:].T),vmin=0, vmax=1)
plt.xlim([0, esn.reservoir_state[:50, 1:].shape[0]])
plt.ylim([0, esn.reservoir_state.shape[1] - 1])
plt.xlabel('n')
plt.ylabel('R[n]')
plt.colorbar(im)
plt.grid()




#!/usr/bin/env python
# coding: utf-8

# # Multipitch tracking using Echo State Networks
# 
# ## Introduction
# 
# In this notebook, we demonstrate how the ESN can deal with multipitch tracking, a challenging multilabel classification problem in music analysis.
# 
# As this is a computationally expensive task, we have pre-trained models to serve as an entry point.
# 
# At first, we import all packages required for this task. You can find the import statements below.

# In[1]:


import numpy as np
import os
import csv
from intervaltree import IntervalTree
from sklearn.base import clone
from joblib import dump, load

from matplotlib import pyplot as plt
plt.rcParams['image.cmap'] = 'jet'
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42
get_ipython().run_line_magic('matplotlib', 'inline')

from IPython.display import set_matplotlib_formats
set_matplotlib_formats('png', 'pdf')

import soundfile as sf
from madmom.processors import SequentialProcessor, ParallelProcessor
from madmom.audio import SignalProcessor, FramedSignalProcessor
from madmom.audio.stft import ShortTimeFourierTransformProcessor
from madmom.audio.filters import LogarithmicFilterbank
from madmom.audio.spectrogram import FilteredSpectrogramProcessor, LogarithmicSpectrogramProcessor, SpectrogramDifferenceProcessor

from pyrcn.echo_state_network import ESNRegressor


# ## Dataset
# 
# For training and evaluation, we have used the MusicNet dataset, which can be downloaded for free from ([https://homes.cs.washington.edu/~thickstn/musicnet.html](https://homes.cs.washington.edu/~thickstn/musicnet.html)). 
# 
# In the following, we assume that you have downloaded and extracted the dataset successfully to the subfolder "dataset/MusicNet".
# 
# Let us define a function to return the names of all audio files from the dataset:
# 

# In[2]:


def get_dataset(dataset_path: str = r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\MusicNet"):
    """
    dataset_path: str: /full/path/to/dataset/
    """
    dataset_path = os.path.normpath(dataset_path)
    train_ids = os.listdir(os.path.join(dataset_path, 'train_data'))
    test_ids = os.listdir(os.path.join(dataset_path, 'test_data'))
    return train_ids, test_ids


# ## Prepare an ESN model
# 
# We have already trained ESN models with different reservoir sizes in the uni- and bidirectional mode. 
# 
# In case you would like to train new models, we provide the code here. Note that this will take a lot of time as the MusicNet is a large-scale dataset.

# In[3]:


try:
    esn = load(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\MusicNet\models\esn_500_True.joblib")
except:
    esn = ESNRegressor(k_in=10, input_scaling=0.2, spectral_radius=0.7, bias=0.7, leakage=0.3, reservoir_size=500, k_res=10,
                       reservoir_activation='tanh', bi_directional=True, solver='ridge', beta=1e-4, random_state=1)


# ## Feature extraction
# 
# The acoustic features extracted from the input signal are obtained by filtering short-term spectra (window length 4096 samples and hop size 10 ms) with a bank of triangular filters in the frequency domain with log-spaced frequencies. The frequency range was 30 Hz to 17 000 Hz and we used 12 filters per octave. We used logarithmic magnitudes and added 1 inside the logarithm to ensure a minimum value of 0 for a frame without energy. The first derivative between adjacent frames was added in order to enrich the features by temporal information. Binary labels indicating absent (value 0) or present (value 1) pitches for each frame are assigned to each frame. Note that this task is a multilabel classification. Each MIDI pitch is a separate class, and multiple or no classes can be active at a discrete frame index.
# 
# For a more detailed description, please have a look in our repository ([https://github.com/TUD-STKS/Automatic-Music-Transcription](https://github.com/TUD-STKS/Automatic-Music-Transcription)) with several detailed examples for music analysis tasks.

# In[4]:


def extract_features(in_folder, file_name):
    full_path_to_audio = os.path.join(in_folder, file_name)
    full_path_to_label = str.replace(full_path_to_audio, 'wav', 'csv').replace('_data', '_labels')
    y, sr = sf.read(file=full_path_to_audio)
    if y.ndim > 1:
        y = y[:, 0]
    y = y / np.max(np.abs(y))

    sig = SignalProcessor(num_channels=1, sample_rate=sr)
    multi = ParallelProcessor([])
    frames = FramedSignalProcessor(frame_size=4096, fps=100)
    stft = ShortTimeFourierTransformProcessor()  # caching FFT window
    filt = FilteredSpectrogramProcessor(filterbank=LogarithmicFilterbank, num_bands=12, fmin=30, fmax=17000, norm_filters=True, unique_filters=True)
    spec = LogarithmicSpectrogramProcessor(log=np.log10, mul=5, add=1.0)
    diff = SpectrogramDifferenceProcessor(diff_ratio=0.5, positive_diffs=True, stack_diffs=np.hstack)
    # process each frame size with spec and diff sequentially
    multi.append(SequentialProcessor([frames, stft, filt, spec, diff]))  #
    pre_processor = SequentialProcessor([sig, multi, np.hstack])

    S = pre_processor.process(data=y) - 1.0
    
    tree = IntervalTree()
    with open(full_path_to_label, 'r') as f:
        reader = csv.DictReader(f, delimiter=',')
        for label in reader:
            start_time = float(label['start_time']) / 44100.
            end_time = float(label['end_time']) / 44100.
            note = int(label['note'])
            tree[start_time:end_time] = (note)

    labels = tree
    y_notes = np.zeros(shape=(S.shape[0], 128), dtype=int)
    for interval in labels.all_intervals:
        y_notes[int(100*interval.begin):int(100*interval.end), int(interval.data)] = 1
    return S, y_notes


# ## Train a model
# 
# This might require a huge amount of time and memory. We have deactivated it for now, because we already offer pre-trained models.
# 
# The pipeline is clear: 
# 
# - Extract features and labels
# - Pass features and labels through ESN
# - Compute output weights and clear no more required helper variables.
# 
# Notice that you can pass the option "update_output_weights=False" to the call to partial_fit. If this option is passed, no output weights are computed after passing the sequence through the network. This is computationally more efficient.
# 
# After passing the entire sequence through the ESN, we need to call finalize() in order to compute the update weights.
# 
# By default, "update_output_weights=False" and it is not necessary to call finalize().

# In[5]:


should_train = False

if should_train:
    train_ids, test_ids = get_dataset()
    for fid in train_ids:
        X, y_true = extract_features(os.path.join(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\MusicNet", 'train_data'), file_name=fid)
        esn.partial_fit(X=X, y=y_true, update_output_weights=False)
    esn.finalize()


# ## Validate the ESN model
# 
# This might require a huge amount of time if a lot of audio files need to be analyzed. 
# 
# The pipeline is similar as before: 
# 
# - Extract features and labels
# - Pass features through ESN and compute outputs
# - Binarize the outputs by simply thresholding
# - Save labels, outputs, binarized outputs for further processing, e.g. visualization or evaluation.

# In[6]:


train_ids, test_ids = get_dataset()
X_train = []
Y_pred_train = []
Y_pred_bin_train = []
Y_true_train = []
for fid in train_ids:
    X, y_true = extract_features(os.path.join(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\MusicNet", 'train_data'), file_name=fid)
    X_train.append(X)
    y_pred = esn.predict(X=X)
    Y_true_train.append(y_true)
    Y_pred_train.append(y_pred)
    Y_pred_bin_train.append(np.asarray(y_pred > 0.3, dtype=int))

X_test = []
Y_pred_test = []
Y_true_test = []
for fid in test_ids:
    X, y_true = extract_features(os.path.join(r"C:\Users\Steiner\Documents\Python\PyRCN\examples\dataset\MusicNet", 'test_data'), file_name=fid)
    X_test.append(X)
    y_pred = esn.predict(X=X)
    Y_true_test.append(y_true)
    Y_pred_test.append(np.asarray(y_pred > 0.3, dtype=int))


# ## Visualization
# 
# We visualize inputs, labels, raw outputs and thresholded outputs in four plots, respectively. 

# In[7]:


fig, axs = plt.subplots(2, 2, figsize=(12, 6))

axs[0, 0].imshow(Y_true_train[0].T, origin='lower', vmin=0, vmax=1, cmap='gray', aspect='auto')
axs[0, 0].set_xlabel('n')
axs[0, 0].set_ylabel('Y_true[n]')
axs[0, 0].grid()
axs[0, 0].set_ylim([21, 108])

axs[0, 1].imshow(X_train[0].T, origin='lower', vmin=0, vmax=1, aspect='auto')
axs[0, 1].set_xlabel('n')
axs[0, 1].set_ylabel('X[n]')
axs[0, 1].grid()

im = axs[1, 0].imshow(Y_pred_train[0].T, origin='lower', aspect='auto')
axs[1, 0].set_xlabel('n')
axs[1, 0].set_ylabel('Y_pred[n]')
axs[1, 0].grid()
axs[1, 0].set_ylim([21, 108])
# fig.colorbar(im, ax=axs[1, 0])

axs[1, 1].imshow(Y_pred_bin_train[0].T, origin='lower', vmin=0, vmax=1, cmap='gray', aspect='auto')
axs[1, 1].set_xlabel('n')
axs[1, 1].set_ylabel('Y_pred_bin[n]')
axs[1, 1].grid()
axs[1, 1].set_ylim([21, 108])

plt.tight_layout()




import os
import glob
import numpy as np
from tqdm import tqdm
import time

from sklearn.base import clone
from sklearn.model_selection import train_test_split, ParameterGrid
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.metrics import mean_squared_error, zero_one_loss
from sklearn.cluster import MiniBatchKMeans, KMeans
# from sklearn_extra.cluster import KMedoids
from sklearn.manifold import TSNE
from joblib import Parallel, delayed, dump, load
from pyrcn.echo_state_network import ESNRegressor
import matplotlib
from matplotlib import pyplot as plt
#Options
params = {'image.cmap' : 'jet',
          'text.usetex' : False,
          'font.size' : 11,
          'axes.titlesize' : 24,
          'axes.labelsize' : 20,
          'lines.linewidth' : 3,
          'lines.markersize' : 10,
          'xtick.labelsize' : 16,
          'ytick.labelsize' : 16,
          'text.latex.unicode': False,
          }
plt.rcParams.update(params)
# plt.rcParams['pdf.fonttype'] = 42
# plt.rcParams['ps.fonttype'] = 42
matplotlib.rcParams['text.usetex'] = False

from IPython.display import set_matplotlib_formats
set_matplotlib_formats('png', 'pdf')

import librosa
import librosa.display

# Print number of files that are included in the dataset
training_files = np.loadtxt(r"Z:\Projekt-Pitch-Datenbank\SPEECH_DATA\SPEECH_DATAsplit\training.txt", dtype=str)
validation_files = np.loadtxt(r"Z:\Projekt-Pitch-Datenbank\SPEECH_DATA\SPEECH_DATAsplit\validation.txt", dtype=str)
test_files = np.loadtxt(r"Z:\Projekt-Pitch-Datenbank\SPEECH_DATA\SPEECH_DATAsplit\test.txt", dtype=str)
print("{0}\t{1}\t{2}".format(len(training_files), len(validation_files), len(test_files)))


# helper function for feature extraction
def extract_features(file_name):
    x, sr = librosa.core.load(file_name, sr=None, mono=False)
    stft_frames = np.abs(librosa.stft(x, n_fft=2048, hop_length=int(0.01*sr), win_length=int(0.04*sr)))**2
    S = librosa.power_to_db(stft_frames, ref=np.max)
    X = S.T
    y = np.zeros(shape=(S.T.shape[0], 2))
    txt_data = np.loadtxt(file_name.replace("MIC", "REF").replace("mic", "ref").replace(".wav", ".f0"), usecols=(0, 1))
    y[2:2+len(txt_data), :] = txt_data
    return X, y


# prepare a scaler to standardize features
scaler = StandardScaler()
try:
    scaler = load("scaler.joblib")
except FileNotFoundError:
    print("Fitting scaler with features from the training set...")
    with tqdm(total=len(training_files)) as pbar:
        for k, file_name in enumerate(training_files):
            X, y = extract_features(file_name)
            scaler.partial_fit(X=X)
            pbar.update(1)
    print("done!")
    dump(scaler, "scaler.joblib")


# Define several error functions for f0 extraction
def gpe(y_true, y_pred):
    """
    Gross pitch error:

    All frames that are considered voiced by both pitch tracker and ground truth,
    for which the relative pitch error is higher than a certain threshold (\SI{20}{\percent}).

    """
    idx = np.nonzero(y_true * y_pred)[0]
    return np.mean(np.abs(y_true[idx] - y_pred[idx]) > 0.2 * y_true[idx])


def vde(y_true, y_pred):
    """
    Voicing Decision Error:

    Proportion of frames for which an incorrect voiced/unvoiced decision is made.

    """
    return zero_one_loss(y_true, y_pred)


def fpe(y_true, y_pred):
    """
    Fine Pitch Error:

    Standard deviation of the distribution of relative error values (in cents) from the frames
    that do not have gross pitch errors
    """
    idx_voiced = np.nonzero(y_true * y_pred)[0]
    idx_correct = np.argwhere(np.abs(y_true - y_pred) <= 0.2 * y_true).ravel()
    idx = np.intersect1d(idx_voiced, idx_correct)
    return 100 * np.std(np.log2(y_pred[idx] / y_true[idx]))


def ffe(y_true, y_pred):
    """
    $f_{0}$ Frame Error:

    Proportion of frames for which an error (either according to the GPE or the VDE criterion) is made.
    FFE can be seen as a single measure for assessing the overall performance of a pitch tracker.
    """
    idx_correct = np.argwhere(np.abs(y_true - y_pred) <= 0.2 * y_true).ravel()
    return 1 - len(idx_correct) / len(y_true)


# Initialize an Echo State Network
base_esn = ESNRegressor(k_in=10, input_scaling=0.1, spectral_radius=0.0, bias=0.0, leakage=1.0, reservoir_size=500,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=0)


# Define a helper function to optimize hyper-parameters
def loss_function(base_esn, params, scaler, training_files, validation_files):
    esn = clone(base_esn).set_params(**params)
    for k, file_name in enumerate(training_files):
        X, y = extract_features(file_name)
        X = scaler.transform(X)
        esn.partial_fit(X=X, y=y, update_output_weights=False)
    esn.finalize()
    ffe_training = [None] * len(training_files)
    for k, file_name in enumerate(training_files):
        X, y = extract_features(file_name)
        X = scaler.transform(X)
        y_pred = esn.predict(X=X)
        ffe_training[k] = ffe(y_true=y[:, 0]*y[:, 1], y_pred=y_pred[:, 0]*(y_pred[:, 1] >= .5))
    ffe_validation = [None] * len(validation_files)
    for k, file_name in enumerate(validation_files):
        X, y = extract_features(file_name)
        X = scaler.transform(X)
        y_pred = esn.predict(X=X)
        ffe_validation[k] = ffe(y_true=y[:, 0]*y[:, 1], y_pred=y_pred[:, 0]*(y_pred[:, 1] >= .5))
    return [np.mean(ffe_training), np.mean(ffe_validation)]


base_esn = ESNRegressor(k_in=10, input_scaling=0.1, spectral_radius=0.9, bias=0.0, leakage=1.0, reservoir_size=500,
                        k_res=10, reservoir_activation='tanh', teacher_scaling=1.0, teacher_shift=0.0,
                        bi_directional=False, solver='ridge', beta=1e-3, random_state=0)

esn = clone(base_esn)
for k, file_name in enumerate(training_files):
    X, y = extract_features(file_name)
    X = scaler.transform(X)
    esn.partial_fit(X=X, y=y, update_output_weights=False)
esn.finalize()

ffe_training = [None] * len(training_files)
for k, file_name in enumerate(training_files):
    X, y = extract_features(file_name)
    X = scaler.transform(X)
    y_pred = esn.predict(X=X)
    ffe_training[k] = ffe(y_true=y[:, 0] * y[:, 1], y_pred=y_pred[:, 0] * (y_pred[:, 1] >= .5))
ffe_validation = [None] * len(validation_files)
for k, file_name in enumerate(validation_files):
    X, y = extract_features(file_name)
    X = scaler.transform(X)
    y_pred = esn.predict(X=X)
    ffe_validation[k] = ffe(y_true=y[:, 0] * y[:, 1], y_pred=y_pred[:, 0] * (y_pred[:, 1] >= .5))



import scipy
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.neural_network._base import ACTIVATIONS
from sklearn.utils import check_random_state
from sklearn.utils import check_X_y, column_or_1d, check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils.multiclass import _check_partial_fit_first_call
from sklearn.exceptions import NotFittedError

from joblib import Parallel, delayed

if scipy.__version__ == '0.9.0' or scipy.__version__ == '0.10.1':
    from scipy.sparse.linalg import eigs as eigens
    from scipy.sparse.linalg import ArpackNoConvergence
else:
    from scipy.sparse.linalg.eigen.arpack import eigs as eigens
    from scipy.sparse.linalg.eigen.arpack import ArpackNoConvergence

_OFFLINE_SOLVERS = ['pinv', 'ridge', 'lasso']


class BaseEchoStateNetwork(BaseEstimator):
    """Base class for ESN classification and regression.

    Warning: This class should not be used directly.
    Use derived classes instead.

    .. versionadded:: 0.00
    """

    def __init__(self, k_in: int = 2, input_scaling: float = 1., spectral_radius: float = 0., bias: float = 0.,
                 ext_bias: int = 0, leakage: float = 1., reservoir_size: int = 500, k_res: int = 10, wash_out: int = 0,
                 reservoir_activation: str = 'tanh', bi_directional: bool = False, teacher_scaling: float = 1.,
                 teacher_shift: float = 0., solver: str = 'ridge', beta: float = 1e-6, random_state: int = None):
        self.k_in = k_in
        self.input_scaling = input_scaling
        self.spectral_radius = spectral_radius
        self.bias = bias
        self.ext_bias = ext_bias
        self.leakage = leakage
        self.reservoir_size = reservoir_size
        self.k_res = k_res
        self.wash_out = wash_out
        self.reservoir_activation = reservoir_activation
        self.bi_directional = bi_directional
        self.teacher_scaling = teacher_scaling
        self.teacher_shift = teacher_shift
        self.solver = solver
        self.beta = beta
        self.random_state = random_state

    def fit(self, X, y, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).

        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN model.
        """
        self._validate_hyperparameters()
        X, y = self._validate_input(X, y)
        self._initialize(y=y, n_features=X.shape[1])
        return self._fit(X, y, update_output_weights=True, n_jobs=n_jobs)

    def finalize(self, n_jobs=0):
        """
        Finalize the training by solving the linear regression problem and deleting xTx and xTy attributes.

        Parameters
        ----------
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------

        """
        self._finalize(n_jobs=n_jobs)

    def _validate_input(self, X, y):
        """
        Ensure that the input and output is in a proper format and transform it if possible.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).

        Returns
        -------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        """
        X, y = check_X_y(X, y, accept_sparse=False, multi_output=True, y_numeric=True)
        if y.ndim == 2 and y.shape[1] == 1:
            y = column_or_1d(y, warn=True)
        y = self.teacher_scaling * y + self.teacher_shift
        return X, y

    def _validate_hyperparameters(self):
        """
        Validate the hyperparameter. Ensure that the parameter ranges and dimensions are valid.
        Returns
        -------

        """
        if self.reservoir_size <= 0:
            raise ValueError("reservoir_size must be > 0, got %s." % self.reservoir_size)
        if self.input_scaling <= 0:
            raise ValueError("input_scaling must be > 0, got %s." % self.input_scaling)
        if self.k_in <= 1 and self.k_in != -1:
            raise ValueError("k_in must be >= 1 or ==-1, got %s." % self.k_in)
        if self.spectral_radius < 0:
            raise ValueError("spectral_radius must be >= 0, got %s." % self.spectral_radius)
        if self.k_res <= 1 and self.k_res != -1:
            raise ValueError("k_res must be >= 1 or ==-1, got %s." % self.k_res)
        if self.bias < 0:
            raise ValueError("bias must be > 0, got %s." % self.bias)
        if self.leakage > 1 or self.leakage < 0:
            raise ValueError("leakage must be >= 0 and <= 1")
        if self.beta < 0.0:
            raise ValueError("beta must be >= 0, got %s." % self.beta)
        # raise ValueError if not registered
        supported_activations = ('identity', 'logistic', 'tanh', 'relu')
        if self.reservoir_activation not in supported_activations:
            raise ValueError("The reservoir_activation '%s' is not supported. Supported "
                             "activations are %s." % (self.reservoir_activation, supported_activations))
        supported_solvers = _OFFLINE_SOLVERS
        if self.solver not in supported_solvers:
            raise ValueError("The solver %s is not supported. Expected one of: %s" %
                             (self.solver, ", ".join(supported_solvers)))

    def initialize_from_outside(self, y, n_features, input_weights=None, reservoir_weights=None, bias_weights=None):
        """
        Initializes weights and other settings from outside. Note that this is no replacement for set_params(**params).

        Parameters
        ----------
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        n_features : int
            The number of input features, e.g. the second dimension of input matrix X
        input_weights : ndarray of shape (n_feature, n_res)
            Externally input weights to be used
        reservoir_weights : ndarray of shape (n_res, n_res)
            Externally input weights to be used

        Returns
        -------
        """
        self._initialize(y=y, n_features=n_features, input_weights=input_weights, reservoir_weights=reservoir_weights, bias_weights=bias_weights)

    def _initialize(self, y, n_features, input_weights=None, reservoir_weights=None, bias_weights=None):
        """
        Initialize everything for the Echo State Network. Set all attributes, allocate weights.

        Parameters
        ----------
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        n_features : int
            The number of input features, e.g. the second dimension of input matrix X
        input_weights : ndarray of shape (n_feature, n_res)
            Externally input weights to be used
        reservoir_weights : ndarray of shape (n_res, n_res)
            Externally input weights to be used

        Returns
        -------

        """
        self._random_state = check_random_state(self.random_state)
        # Either one- or multi-dimensional output
        if y.ndim == 2:
            self.n_outputs_ = y.shape[1]
        else:
            self.n_outputs_ = 1

        # Initialize number of training samples
        self._n_samples = 0

        # initialize all weights the model consists of
        input_weights_init, reservoir_weights_init, bias_weights_init, output_weights_init = self._init_weights(n_features)
        if input_weights is not None:
            self.input_weights_ = input_weights
        else:
            self.input_weights_ = input_weights_init
        if reservoir_weights is not None:
            self.reservoir_weights_ = reservoir_weights
        else:
            self.reservoir_weights_ = reservoir_weights_init
        if bias_weights is not None:
            self.bias_weights_ = bias_weights
        else:
            self.bias_weights_ = bias_weights_init

        self.output_weights_ = output_weights_init
        self._init_state_collection_matrices()

    def _init_state_collection_matrices(self):
        # collect the mean and variances of all reservoir nodes. This is required for the dropout strategy.
        self._activations_mean = np.zeros(shape=(self.reservoir_size,))
        self._activations_var = np.zeros(shape=(self.reservoir_size,))
        # initialize xTx and xTy for linear regression. Will be deleted after the training is finalized.
        if self.bi_directional:
            self._xTx = np.zeros(shape=(2 * self.reservoir_size + 1, 2 * self.reservoir_size + 1))
            self._xTy = np.zeros(shape=(2 * self.reservoir_size + 1, self.n_outputs_))
        else:
            self._xTx = np.zeros(shape=(self.reservoir_size + 1, self.reservoir_size + 1))
            self._xTy = np.zeros(shape=(self.reservoir_size + 1, self.n_outputs_))

    def _init_weights(self, n_features):
        """
        Initialize all weight matrices, e.g. connections from the input to the reservoir, and recurrent connections
        inside the reservoir.
        Parameters
        ----------
        n_features : int
            The number of input features, e.g. the second dimension of input matrix X

        Returns
        -------

        """
        # Input-to-reservoir weights, drawn from uniform distribution.
        idx_co = 0
        if self.k_in == -1:
            input_weights_init = self._random_state.rand(self.reservoir_size, n_features) * 2 - 1
        else:
            nr_entries = np.int32(self.reservoir_size*self.k_in)
            ij = np.zeros((2, nr_entries), dtype=int)
            data_vec = self._random_state.rand(nr_entries) * 2 - 1
            for en in range(self.reservoir_size):
                per = self._random_state.permutation(n_features)[:self.k_in]
                ij[0][idx_co:idx_co+self.k_in] = en
                ij[1][idx_co:idx_co+self.k_in] = per
                idx_co = idx_co + self.k_in
            input_weights_init = scipy.sparse.csc_matrix((data_vec, ij),
                                                         shape=(self.reservoir_size, n_features), dtype='float64')
        # Recurrent weights inside the reservoir, drawn from a standard normal distribution.
        converged = False
        # Recurrent weights are normalized to a unitary spectral radius if possible.
        attempts = 50
        while not converged and attempts > 0:
            try:
                if self.k_res == -1:
                    reservoir_weights_init = self._random_state.randn(self.reservoir_size, self.reservoir_size)
                else:
                    idx_co = 0
                    nr_entries = np.int32(self.reservoir_size * self.k_res)
                    ij = np.zeros((2, nr_entries), dtype=int)
                    data_vec = self._random_state.randn(nr_entries)
                    for en in range(self.reservoir_size):
                        per = self._random_state.permutation(self.reservoir_size)[:self.k_res]
                        ij[0][idx_co:idx_co + self.k_res] = en
                        ij[1][idx_co:idx_co + self.k_res] = per
                        idx_co += self.k_res

                    reservoir_weights_init = scipy.sparse.csc_matrix((data_vec, ij),
                                                                     shape=(self.reservoir_size, self.reservoir_size),
                                                                     dtype='float64')
                we = eigens(reservoir_weights_init, return_eigenvectors=False, k=6)
                converged = True
            except ArpackNoConvergence:
                print("WARNING: No convergence! Redo {0} times...".format(attempts-1))
                attempts = attempts - 1
                if attempts == 0:
                    print("WARNING: Returning possibly invalid eigenvalues...")
                we = ArpackNoConvergence.eigenvalues
                pass

        reservoir_weights_init *= (1. / np.amax(np.absolute(we)))
        # Bias weights, fully connected bias for the reservoir nodes, drawn from uniform distribution.
        if self.ext_bias > 0:
            bias_weights_init = (self._random_state.rand(self.reservoir_size, self.ext_bias) * 2 - 1)
        else:
            bias_weights_init = (self._random_state.rand(self.reservoir_size) * 2 - 1)
        output_weights_init = None  # np.zeros(shape=(self.reservoir_size + 1, self.n_outputs_))
        return input_weights_init, reservoir_weights_init, bias_weights_init, output_weights_init

    def _fit(self, X, y, incremental=False, update_output_weights=True, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        incremental : bool, default False
            If True, the network can be fitted with data that does not fit into memory. After each call to fit,
            output weights are trained if update_output_weights == True
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN model.
        """
        n_samples, n_features = X.shape
        if self.ext_bias > 0:
            n_features = n_features - self.ext_bias
        # Ensure y is 2D
        if y.ndim == 1:
            y = y.reshape((-1, 1))
        self.n_outputs_ = y.shape[1]
        if (not hasattr(self, 'input_weights_')) or (not hasattr(self, 'reservoir_weights_')) \
                or (not hasattr(self, 'bias_weights_')) or not incremental:
            # First time training the model
            self._initialize(y, n_features)

        # Run the offline optimization solver
        if self.solver in _OFFLINE_SOLVERS:
            self._fit_offline(X, y, incremental, update_output_weights=update_output_weights, n_jobs=n_jobs)
        self.is_fitted_ = True
        return self

    def _pass_through_reservoir(self, X):
        """
        Pass the data forward and, if required, backwards through the reservoir.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        Returns
        -------
        reservoir_state : ndarray of shape (n_samples, reservoir_size)
            The collected reservoir states
        """
        reservoir_state = self._forward_pass(reservoir_inputs=X)
        if self.bi_directional:
            reservoir_state = \
                np.concatenate((reservoir_state, np.flipud(self._forward_pass(reservoir_inputs=np.flipud(X)))), 1)
        reservoir_state = np.concatenate((np.ones((reservoir_state.shape[0], 1)), reservoir_state), 1)
        return reservoir_state

    def _fit_offline(self, X, y, incremental=False, update_output_weights=True, n_jobs: int = 0):
        """
        Do a single fit of the model on the entire dataset passed trough.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        incremental : bool, default False
            If True, the network can be fitted with data that does not fit into memory. After each call to fit,
            output weights are trained if update_output_weights == True
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------

        """
        n_samples = X.shape[0]
        self._n_samples = self._n_samples + n_samples - self.wash_out

        reservoir_state = self._pass_through_reservoir(X=X)

        if incremental:
            self._xTx = self._xTx + np.dot(reservoir_state[self.wash_out:, :].T, reservoir_state[self.wash_out:, :])
            self._xTy = self._xTy + np.dot(reservoir_state[self.wash_out:, :].T, y[self.wash_out:, :])
            if self.bi_directional:
                new_activations_mean = np.mean(reservoir_state[self.wash_out:, :], axis=0)[1:self.reservoir_size + 1]
                new_activations_var = np.var(reservoir_state[self.wash_out:, :], axis=0)[1:self.reservoir_size + 1]
            else:
                new_activations_mean = np.mean(reservoir_state[self.wash_out:, :], axis=0)[1:]
                new_activations_var = np.var(reservoir_state[self.wash_out:, :], axis=0)[1:]
            m = self._n_samples
            n = reservoir_state.shape[0]
            tmp_activations_mean = self._activations_mean
            self._activations_mean = m/(m+n)*tmp_activations_mean + n/(m+n)*new_activations_mean
            self._activations_var = m / (m + n) * self._activations_var + n / (m + n)*new_activations_var + \
                                    m * n / (m + n)**2 * (tmp_activations_mean - new_activations_mean)**2
        else:
            self._xTx = np.dot(reservoir_state[self.wash_out:, :].T, reservoir_state[self.wash_out:, :])
            self._xTy = np.dot(reservoir_state[self.wash_out:, :].T, y[self.wash_out:, :])
            if self.bi_directional:
                self.activations_mean = np.mean(reservoir_state[self.wash_out:, :], axis=0)[1:self.reservoir_size + 1]
                self.activations_var = np.var(reservoir_state[self.wash_out:, :], axis=0)[1:self.reservoir_size + 1]
            else:
                self.activations_mean = np.mean(reservoir_state[self.wash_out:, :], axis=0)[1:]
                self.activations_var = np.var(reservoir_state[self.wash_out:, :], axis=0)[1:]

        if update_output_weights:
            self._compute_output_weights(n_jobs=n_jobs)
        else:
            self.output_weights_ = None

        if not incremental:
            self._xTx = None
            self._xTy = None

    def _forward_pass(self, reservoir_inputs):
        """
        Perform a forward pass on the network by computing the values
        of the neurons in the hidden layers and the output layer.

        Parameters
        ----------
        reservoir_inputs : ndarray of shape (n_samples, n_features)
            The input data

        Returns
        -------

        """
        n_samples, n_features = reservoir_inputs.shape
        reservoir_state = np.zeros(shape=(n_samples+1, self.reservoir_size))
        if self.ext_bias > 0:
            for sample in range(n_samples):
                if scipy.sparse.issparse(self.input_weights_):
                    a = self.input_weights_ * reservoir_inputs[sample, :-self.ext_bias] * self.input_scaling
                else:
                    a = np.dot(self.input_weights_, self.input_scaling * reservoir_inputs[sample, :-self.ext_bias])
                if scipy.sparse.issparse(self.reservoir_weights_):
                    b = self.reservoir_weights_ * reservoir_state[sample, :] * self.spectral_radius
                else:
                    b = np.dot(self.reservoir_weights_, self.spectral_radius * reservoir_state[sample, :])

                reservoir_state[sample + 1, :] = ACTIVATIONS[self.reservoir_activation](
                    np.atleast_2d(a + b).T + self.bias_weights_ * self.bias * reservoir_inputs[sample, -self.ext_bias]).flatten()
                reservoir_state[sample + 1, :] = (1 - self.leakage) * reservoir_state[sample, :]\
                                                 + self.leakage * reservoir_state[sample + 1, :]
        else:
            for sample in range(n_samples):
                if scipy.sparse.issparse(self.input_weights_):
                    a = self.input_weights_ * reservoir_inputs[sample, :] * self.input_scaling
                else:
                    a = np.dot(self.input_weights_, self.input_scaling * reservoir_inputs[sample, :])
                if scipy.sparse.issparse(self.reservoir_weights_):
                    b = self.reservoir_weights_ * reservoir_state[sample, :] * self.spectral_radius
                else:
                    b = np.dot(self.reservoir_weights_, self.spectral_radius * reservoir_state[sample, :])

                reservoir_state[sample+1, :] = \
                    ACTIVATIONS[self.reservoir_activation](a + b + self.bias_weights_*self.bias)
                reservoir_state[sample+1, :] = \
                    (1 - self.leakage) * reservoir_state[sample, :] + self.leakage * reservoir_state[sample+1, :]
        return reservoir_state[1:, :]

    def partial_fit(self, X, y, update_output_weights=True, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y without finalizing it. This can be used to add more training
        data later.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN model.
        """
        if self.solver not in _OFFLINE_SOLVERS:
            raise AttributeError('partial_fit is only available for offline optimizers, not for %s.' % self.solver)
        return self._partial_fit(X=X, y=y, update_output_weights=update_output_weights, n_jobs=n_jobs)

    def _partial_fit(self, X, y, update_output_weights=True, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y without finalizing it. This can be used to add more training
        data later.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN model.
        """
        return self._fit(X, y, incremental=True, update_output_weights=update_output_weights, n_jobs=n_jobs)

    def drop_out(self, drop_out_rate=0.0):
        """
        Experimental dropout strategy for the ESN. After passing some data through the network and collecting reservoir
        activations, a percentile of nodes with very low activity can be removed. Note that this works best without
        bidirectional mode right now.

        Warnings : This is experimental and needs to be validated at first. Publications will follow!!!

        Parameters
        ----------
        drop_out_rate : double, default 0.0
            Determines the percentile of nodes to be removed

        Returns
        -------

        """
        self._drop_out(drop_out_rate=drop_out_rate)

    def _drop_out(self, drop_out_rate):
        """
        Experimental dropout strategy for the ESN. After passing some data through the network and collecting reservoir
        activations, a percentile of nodes with very low activity can be removed. Note that this works best without
        bidirectional mode right now.

        Warnings : This is experimental and needs to be validated at first. Publications will follow!!!

        Warnings : After dropout, the output weights need to be trained again, because several nodes have been cut off!

        Parameters
        ----------
        drop_out_rate : double, default 0.0
            Determines the percentile of nodes to be removed

        Returns
        -------

        """
        if drop_out_rate > 0.0:
            new_reservoir_size = int(drop_out_rate * self.reservoir_size)
            idx_to_drop_ = np.argsort(self._activations_var)[::-1][int(drop_out_rate * self.reservoir_size):]
            self.bias_weights_ = np.delete(self.bias_weights_, idx_to_drop_)
            self.input_weights_ = scipy.sparse.csc_matrix(
                np.delete(self.input_weights_.toarray(), idx_to_drop_, axis=0), dtype='float64')
            self.reservoir_weights_ = scipy.sparse.csc_matrix(
                np.delete(np.delete(self.reservoir_weights_.toarray(), idx_to_drop_, axis=0), idx_to_drop_, axis=1),
                dtype='float64')

            self._n_samples = 0

            self.reservoir_size = new_reservoir_size
            self._init_state_collection_matrices()

    def _finalize(self, n_jobs: int = 0):
        """
        This finalizes the training of a model. No more required attributes, such as activations, xTx, xTy will be
        removed.

        Warnings : The model cannot be improved with additional data afterwards!!!

        Parameters
        ----------
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------

        """
        if self.output_weights_ is None:
            self._compute_output_weights(n_jobs=n_jobs)

        self._xTx = None
        self._xTy = None
        self._activations_mean = None
        self._activations_var = None
        self.is_fitted_ = True

    def _compute_output_weights(self, n_jobs=0):
        """
        This is a helper function to compute the output weights using linear regression
        Parameters
        ----------
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------

        """
        if self.solver == 'pinv':
            inv_xTx = np.linalg.inv(self._xTx)
        elif self.solver == 'ridge':
            lmda = self.beta ** 2 * self._n_samples
            inv_xTx = np.linalg.inv(self._xTx + lmda * np.eye(self._xTx.shape[0]))
        else:
            print("Warning: Not implemented. Falling back to pinv solution")
            inv_xTx = np.linalg.inv(self._xTx)
        if n_jobs > 0:
            self.output_weights_ = Parallel(n_jobs=n_jobs)(
                delayed(np.dot)(inv_xTx, self._xTy[:, n]) for n in range(self.n_outputs_))
        else:
            self.output_weights_ = np.dot(inv_xTx, self._xTy)

    def predict(self, X, keep_reservoir_state=False):
        """
        Predict using the trained ESN model

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input data.
        keep_reservoir_state : bool, default False
            If True, the reservoir state is kept and can be accessed from outside. This is useful for visualization
        Returns
        -------
        y_pred : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The predicted values
        """
        check_is_fitted(self, ['input_weights_', 'reservoir_weights_', 'bias_weights_', 'output_weights_'])
        if not self.output_weights_.any():
            msg = ("This %(name)s instance is not fitted yet. Call 'fit' with "
                   "appropriate arguments before using this method.")
            raise NotFittedError(msg % {'name': type(self).__name__})
        X = check_array(X, accept_sparse=False)
        y_pred = self._predict(X=X, keep_reservoir_state=keep_reservoir_state)
        return y_pred

    def _predict(self, X, keep_reservoir_state=False):
        """
        Predict using the trained ESN model

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input data.
        keep_reservoir_state : bool, default False
            If True, the reservoir state is kept and can be accessed from outside. This is useful for visualization
        Returns
        -------
        y_pred : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The predicted values
        """
        reservoir_state = self._pass_through_reservoir(X=X)
        if keep_reservoir_state:
            self.reservoir_state = reservoir_state
        y_pred = safe_sparse_dot(reservoir_state, self.output_weights_)
        return y_pred


class ESNClassifier(BaseEchoStateNetwork, ClassifierMixin):
    """
    Echo State Network classifier.

    This model optimizes the mean squared error loss function using linear regression.

    .. versionadded:: 0.00

    Parameters
    ----------
    k_in : int, default 2
        This element represents the sparsity of the connections between the input and recurrent nodes.
        It determines the number of features that every node inside the reservoir receives.
    input_scaling : float, default 1.0
        This element represents the input scaling factor from the input to the reservoir. It is a global scaling factor
        for the input weight matrix.
    spectral_radius : float, default 0.0
        This element represents the spectral radius of the reservoir weights. It is a global scaling factor for the
        reservoir weight matrix.
    bias : float, default 0.0
        This element represents the bias scaling of the bias weights. It is a global scaling factor for the bias weight
        matrix.
    ext_bias : int, default 0
        This element represents the number of external biases to be used. If greater than zero, they are supposed to be
        appended to the input matrix.
    leakage : float, default 1.0
        This element represents the leakage of the reservoir. Depending on the value, it acts as a short- or long-term
        memory coefficient.
    reservoir_size : int, default 500
        This element represents the number of neurons in the reservoir.
    k_res : int, default 10
        This element represents the sparsity of the connections inside the reservoir. It determines the number of nodes
        that every node inside the reservoir is connected with.
    wash_out: int, default 0
        This element defines the number of first samples to be not regarded for training.
    reservoir_activation : {'tanh', 'identity', 'logistic', 'relu'}
        This element represents the activation function in the reservoir.
            - 'identity', no-op activation, useful to implement linear bottleneck, returns f(x) = x
            - 'logistic', the logistic sigmoid function, returns f(x) = 1 / (1 + exp(-x)).
            - 'tanh', the hyperbolic tan function, returns f(x) = tanh(x).
            - 'relu', the rectified linear unit function, returns f(x) = max(0, x)
    teacher_scaling : float, default 1.0
        This element represents the teacher scaling factor. In some cases, for example for imbalanced datasets, it might
        make sense to increase this hyperparameter.
    teacher_shift : float, default 0.0
        This element represents the teacher shift factor. In some cases, for example for imbalanced datasets, it might
        make sense to increase this hyperparameter.
    bi_directional : bool, default False
        If True, the input sequences are passed through the network two times, forward and backward.
    solver : {'ridge', 'pinv'}
        The solver for weight optimization.
        - 'pinv' uses the pseudoinverse solution of linear regression.
        - 'ridge' uses L2 penalty while computing the linear regression
    beta : float, optional, default 0.0001
        L2 penalty (regularization term) parameter.
    random_state : int, RandomState instance or None, optional, default None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Attributes
    ----------
    TODO

    Notes
    -----
    TODO

    References
    ----------
    TODO
    """
    def __init__(self, k_in: int = 2, input_scaling: float = 1., spectral_radius: float = 0., bias: float = 0.,
                 ext_bias: int = 0, leakage: float = 1., reservoir_size: int = 500, k_res: int = 10, wash_out: int = 0,
                 reservoir_activation: str = 'tanh', bi_directional: bool = False, teacher_scaling: float = 1.,
                 teacher_shift: float = 0., solver: str = 'ridge', beta: float = 1e-6, random_state: int = None):
        super().__init__(k_in=k_in, input_scaling=input_scaling, spectral_radius=spectral_radius, bias=bias,
                         ext_bias=ext_bias, leakage=leakage, reservoir_size=reservoir_size, k_res=k_res,
                         wash_out=wash_out, reservoir_activation=reservoir_activation, bi_directional=bi_directional,
                         teacher_scaling=teacher_scaling, teacher_shift=teacher_shift, solver=solver, beta=beta,
                         random_state=random_state)

    def _validate_input(self, X, y):
        """
        Ensure that the input and output is in a proper format and transform it if possible.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).

        Returns
        -------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        """
        X, y = check_X_y(X, y, accept_sparse=False, multi_output=True)
        if y.ndim == 2 and y.shape[1] == 1:
            y = column_or_1d(y, warn=True)

        self._label_binarizer = LabelBinarizer()
        self._label_binarizer.fit(y)
        self.classes_ = self._label_binarizer.classes_
        y = self._label_binarizer.transform(y)
        y = self.teacher_scaling * y + self.teacher_shift
        return X, y

    def fit(self, X, y, n_jobs: int = 0):
        """
        Fit the model to the data matrix X and target(s) y.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN model.
        """
        self._validate_hyperparameters()
        X, y = self._validate_input(X, y)
        if self.ext_bias:
            self._initialize(y=y, n_features=X.shape[1] - 1)
        else:
            self._initialize(y=y, n_features=X.shape[1])
        return self._fit(X, y, incremental=False, update_output_weights=True, n_jobs=n_jobs)

    def predict(self, X, keep_reservoir_state=False):
        """
        Predict the classes using the trained ESN classifier

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input data.
        keep_reservoir_state : bool, default False
            If True, the reservoir state is kept and can be accessed from outside. This is useful for visualization
        Returns
        -------
        y_pred : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The predicted classes
        """
        check_is_fitted(self, ['input_weights_', 'reservoir_weights_', 'bias_weights_', 'output_weights_'])
        if self.output_weights_.size == 0:
            msg = ("This %(name)s instance is not fitted yet. Call 'fit' with "
                   "appropriate arguments before using this method.")
            raise NotFittedError(msg % {'name': type(self).__name__})
        y_pred = super().predict(X, keep_reservoir_state=keep_reservoir_state)

        if self.n_outputs_ == 1:
            y_pred = y_pred.ravel()
        return self._label_binarizer.inverse_transform(y_pred)

    def partial_fit(self, X, y, classes=None, update_output_weights=True, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y without finalizing it. This can be used to add more training
        data later.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        classes : ndarray of shape (class labels, )
            The class labels to be predicted!
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN classifier.
        """
        if self.solver not in _OFFLINE_SOLVERS:
            raise AttributeError('partial_fit is only available for offline optimizers, not for %s.' % self.solver)
        return self._partial_fit(X=X, y=y, classes=classes, update_output_weights=update_output_weights, n_jobs=n_jobs)

    def _partial_fit(self, X, y, classes=None, update_output_weights=True, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y without finalizing it. This can be used to add more training
        data later.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN classifier.
        """
        if _check_partial_fit_first_call(self, classes):
            self._label_binarizer = LabelBinarizer().fit(classes)
            if self.ext_bias:
                super()._initialize(y=y, n_features=X.shape[1] - 1)
            else:
                super()._initialize(y=y, n_features=X.shape[1])

        super()._partial_fit(X, y, update_output_weights=update_output_weights, n_jobs=n_jobs)
        return self

    def predict_proba(self, X, keep_reservoir_state=False):
        """
        Predict the probability estimates using the trained ESN classifier

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input data.
        keep_reservoir_state : bool, default False
            If True, the reservoir state is kept and can be accessed from outside. This is useful for visualization
        Returns
        -------
        y_pred : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The predicted probability estimates
        """
        y_pred = super().predict(X, keep_reservoir_state=keep_reservoir_state)
        y_pred = np.maximum(y_pred, 1e-3)

        if self.n_outputs_ == 1:
            y_pred = y_pred.ravel()

        if y_pred.ndim == 1:
            return np.vstack([1 - y_pred, y_pred]).T
        else:
            return y_pred

    def predict_log_proba(self, X, keep_reservoir_state=False):
        """
        Predict the logarithmic probability estimates using the trained ESN classifier

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input data.
        keep_reservoir_state : bool, default False
            If True, the reservoir state is kept and can be accessed from outside. This is useful for visualization
        Returns
        -------
        y_pred : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The predicted logarithmic probability estimates
        """
        y_pred = self.predict_proba(X=X, keep_reservoir_state=keep_reservoir_state)
        return np.log(y_pred)


class ESNRegressor(BaseEchoStateNetwork, RegressorMixin):
    """
    Echo State Network regressor.

    This model optimizes the mean squared error loss function using linear regression.

    .. versionadded:: 0.00

    Parameters
    ----------
    k_in : int, default 2
        This element represents the sparsity of the connections between the input and recurrent nodes.
        It determines the number of features that every node inside the reservoir receives.
    input_scaling : float, default 1.0
        This element represents the input scaling factor from the input to the reservoir. It is a global scaling factor
        for the input weight matrix.
    spectral_radius : float, default 0.0
        This element represents the spectral radius of the reservoir weights. It is a global scaling factor for the
        reservoir weight matrix.
    bias : float, default 0.0
        This element represents the bias scaling of the bias weights. It is a global scaling factor for the bias weight
        matrix.
    ext_bias : int, default 0
        This element represents the number of external biases to be used. If greater than zero, they are supposed to be
        appended to the input matrix.
    leakage : float, default 1.0
        This element represents the leakage of the reservoir. Depending on the value, it acts as a short- or long-term
        memory coefficient.
    reservoir_size : int, default 500
        This element represents the number of neurons in the reservoir.
    k_res : int, default 10
        This element represents the sparsity of the connections inside the reservoir. It determines the number of nodes
        that every node inside the reservoir is connected with.
    wash_out: int, default 0
        This element defines the number of first samples to be not regarded for training.
    reservoir_activation : {'tanh', 'identity', 'logistic', 'relu'}
        This element represents the activation function in the reservoir.
            - 'identity', no-op activation, useful to implement linear bottleneck, returns f(x) = x
            - 'logistic', the logistic sigmoid function, returns f(x) = 1 / (1 + exp(-x)).
            - 'tanh', the hyperbolic tan function, returns f(x) = tanh(x).
            - 'relu', the rectified linear unit function, returns f(x) = max(0, x)
    teacher_scaling : float, default 1.0
        This element represents the teacher scaling factor. In some cases, for example for imbalanced datasets, it might
        make sense to increase this hyperparameter.
    teacher_shift : float, default 0.0
        This element represents the teacher shift factor. In some cases, for example for imbalanced datasets, it might
        make sense to increase this hyperparameter.
    bi_directional : bool, default False
        If True, the input sequences are passed through the network two times, forward and backward.
    solver : {'ridge', 'pinv'}
        The solver for weight optimization.
        - 'pinv' uses the pseudoinverse solution of linear regression.
        - 'ridge' uses L2 penalty while computing the linear regression
    beta : float, optional, default 0.0001
        L2 penalty (regularization term) parameter.
    random_state : int, RandomState instance or None, optional, default None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Attributes
    ----------
    TODO

    Notes
    -----
    TODO

    References
    -----------
    TODO
    """
    def __init__(self, k_in: int = 2, input_scaling: float = 1., spectral_radius: float = 0., bias: float = 0.,
                 ext_bias: int = 0, leakage: float = 1., reservoir_size: int = 500, k_res: int = 10, wash_out: int = 0,
                 reservoir_activation: str = 'tanh', bi_directional: bool = False, teacher_scaling: float = 1.,
                 teacher_shift: float = 0., solver: str = 'ridge', beta: float = 1e-6, random_state: int = None):
        super().__init__(k_in=k_in, input_scaling=input_scaling, spectral_radius=spectral_radius, bias=bias,
                         ext_bias=ext_bias, leakage=leakage, reservoir_size=reservoir_size, k_res=k_res,
                         wash_out=wash_out, reservoir_activation=reservoir_activation, bi_directional=bi_directional,
                         teacher_scaling=teacher_scaling, teacher_shift=teacher_shift, solver=solver, beta=beta,
                         random_state=random_state)

    def fit(self, X, y, n_jobs=0):
        self._validate_hyperparameters()
        X, y = self._validate_input(X, y)
        self._initialize(y=y, n_features=X.shape[1])
        return self._fit(X, y, update_output_weights=True, n_jobs=n_jobs)

    def predict(self, X, keep_reservoir_state=False):
        """
        Predict the classes using the trained ESN regressor

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input data.
        keep_reservoir_state : bool, default False
            If True, the reservoir state is kept and can be accessed from outside. This is useful for visualization
        Returns
        -------
        y_pred : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The predicted classes
        """
        y_pred = super().predict(X, keep_reservoir_state=keep_reservoir_state)

        if self.n_outputs_ == 1:
            y_pred = y_pred.ravel()

        return y_pred

    def partial_fit(self, X, y, update_output_weights=True, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y without finalizing it. This can be used to add more training
        data later.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN classifier.
        """
        if self.solver not in _OFFLINE_SOLVERS:
            raise AttributeError("partial_fit is only available for offline optimizer. %s is not offline"
                                 % self.solver)
        return self._partial_fit(X=X, y=y, update_output_weights=update_output_weights, n_jobs=n_jobs)

    def _partial_fit(self, X, y, update_output_weights=True, n_jobs=0):
        """
        Fit the model to the data matrix X and target(s) y without finalizing it. This can be used to add more training
        data later.
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The input data
        y : ndarray of shape (n_samples, ) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in regression).
        update_output_weights : bool, default True
            If False, no output weights are computed after passing the current data through the network.
            This is computationally more efficient in case of a lot of outputs and a large dataset that is fitted
            incrementally.
        n_jobs : int, default: 0
            If n_jobs is larger than 1, then the linear regression for each output dimension is computed separately
            using joblib.

        Returns
        -------
        self : returns a trained ESN classifier.
        """
        super()._partial_fit(X, y, update_output_weights=update_output_weights, n_jobs=n_jobs)
        return self



#!/usr/bin/env python
# coding: utf-8

# # Stock Price Prediction using Echo State Networks
# 
# ## Introduction
# 
# Time Series Prediction is one important regression task, which can be solved using several machine learning techniques.
# 
# In this notebook, we briefly introduce one very basic example for time series prediction:
# 
# Stock Price Prediction
# 
# Disclaimer: We are signal processing experts, not financial advisors. Do not use any of the models presented herein to steer your investments.
# 
# At first, we need to import all required packages

# In[1]:


import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error

from matplotlib import pyplot as plt
plt.rcParams['image.cmap'] = 'jet'
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42
# get_ipython().run_line_magic('matplotlib', 'inline')

from IPython.display import set_matplotlib_formats
set_matplotlib_formats('png', 'pdf')

from pyrcn.echo_state_network import ESNRegressor


# ## Loading and visualizing stock prices
# 
# A good ressource for stock prices is Yahoo Finance ([https://finance.yahoo.com/](https://finance.yahoo.com/)), where a
# lot of financial data can be downloaded as csv files.
# 
# We have already downloaded several stock prices, which are all stored in the directory "dataset".
# 
# Now, we are working with the gold price in USD (Gold Aug 20(GC=F)) between 2000-02-28 and 2020-05-28.

# In[2]:


time_series = pd.read_csv(filepath_or_buffer="./dataset/GC=F.csv")
time_series.head()


# Analyzing the dataset, we can analyze the structure. The first column is the date, the second column is the open value,
# e.g. the first value of a day.
# The third column is the highest value of a specific day, the fourth column the lowest and "Close" is the final value of
# a day. The "Adj Close" is a corrected  final value, I DON 'T YET KNOW ANYTHING ABOUT VOLUME!!!
# 
# Here, we just use the "Close" value, the final value of one day. 

# In[3]:


time_series = time_series.drop(columns=["Open", "High", "Low", "Adj Close", "Volume"], axis=1)
time_series = time_series.dropna()
time_series.head()


# As one can see, the stock price trends upwards for the first few thousand days. Then goes down at around $n=3300$ before climbing again after around $n = 4900$. 

# In[4]:


plt.figure()
time_series.plot(x='Date', y='Close', legend=False, grid=True)
plt.xlabel("Timestamp")
plt.ylabel("Price")


# We pre-processed the dataset by removing undefined values, namely, weekends and public holidays. The remaining values were normalized to be in a range of $[0 1]$.

# In[5]:


prices = np.asarray(time_series['Close'])

scaler = MinMaxScaler()
prices = scaler.fit_transform(X=prices.reshape(-1, 1))


# Define Train/Test lengths:
# 
# The first 3000 values were used to train the ESN and to optimize the hyper-parameters and the rest for testing. 

# In[6]:


train_len = 3000
future_len = 1
future_total = len(prices) - train_len


# Echo State Network preparation

# In[7]:


esn = ESNRegressor(k_in=1, input_scaling=0.6, spectral_radius=0.9, bias=0.0, ext_bias=False, leakage=1.0,
                   reservoir_size=100, k_res=10, reservoir_activation='tanh', bi_directional=False,
                   teacher_scaling=1., teacher_shift=0., solver='ridge', beta=1e-8, random_state=0)


# Training and Prediction.

# In[8]:


train_in = prices[0:train_len, :]
train_out = prices[0+1:train_len+1, :]
test_in = prices[0:train_len+future_total - future_len, :]
test_out = prices[future_len:train_len+future_total, :]

esn.fit(X=train_in, y=train_out.ravel())
train_pred = esn.predict(X=train_in)
test_pred = esn.predict(X=test_in)

train_err = mean_squared_error(y_true=train_out, y_pred=train_pred)
test_err = mean_squared_error(y_true=test_out[train_len+1:, :], y_pred=test_pred[train_len+1:])

print("Train MSE:\t{0}".format(train_err))
print("Test MSE:\t{0}".format(test_err))


# We see that the ESN even captures the downward trend in the test set, although it has not seen any longer downward movement during the training.

# In[9]:


prices_pred = np.empty_like(prices)
prices_pred[:] = np.nan
prices_pred[train_len + future_len:, 0] = test_pred[train_len:]

plt.figure()
plt.plot(scaler.inverse_transform(test_out), label='Target')
plt.plot(scaler.inverse_transform(prices_pred), label='Predicted')
plt.xlabel("Timestamp")
plt.ylabel("Price")
plt.legend()


# Disclaimer: We are signal processing experts, not financial advisors. Do not use any of the models presented herein to steer your investments.



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PyRCN",
    version="0.0.4b",
    author="Peter Steiner",
    author_email="peter.steiner@tu-dresden.de",
    description="A Python3 framework for Reservoir Computing with a scikit-learn-compatible API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TUD-STKS/PyRCN",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
    ],
    keywords='PyRCN, Echo State Network',
    project_urls={
        'Documentation': 'https://pyrcn.readthedocs.io/',
        'Funding': 'https://pyrcn.readthedocs.io/',
        'Source': 'https://github.com/TUD-STKS/PyRCN',
        'Tracker': 'https://github.com/TUD-STKS/PyRCN/issues',
    },
    install_requires=[
        'scikit-learn>=0.22.1',
        'numpy>=1.18.1',
        'scipy>=1.2.0',
        'joblib>=0.13.2',
    ],
    python_requires='>=3.6',
)



#!/usr/bin/env python
# coding: utf-8

# # Recognizing hand-written digits
# 
# ## Introduction
# 
# This notebook adapts the existing example of applying support vector classification from scikit-learn ([https://scikit-learn.org/stable/auto_examples/classification/plot_digits_classification.html#sphx-glr-auto-examples-classification-plot-digits-classification-py](https://scikit-learn.org/stable/auto_examples/classification/plot_digits_classification.html#sphx-glr-auto-examples-classification-plot-digits-classification-py)) to PyRCN to demonstrate, how PyRCN can be used to classify hand-written digits.
# 
# The tutorial is based on numpy, scikit-learn and PyRCN. We are using the ESNRegressor, because we further process the outputs of the ESN. Note that the same can also be done using the ESNClassifier. Then, during prediction, we simply call "predict_proba".
# 
# This tutorial requires the Python modules numpy, scikit-learn, matplotlib and pyrcn.

# In[ ]:


import numpy as np
from sklearn.datasets import load_digits
from sklearn.model_selection import train_test_split
from sklearn.model_selection import ParameterGrid
from sklearn.preprocessing import OneHotEncoder
from sklearn.base import clone
from sklearn.metrics import mean_squared_error, classification_report, confusion_matrix, ConfusionMatrixDisplay
from joblib import Parallel, delayed
import time

from matplotlib import pyplot as plt
plt.rcParams['image.cmap'] = 'jet'
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42

from pyrcn.echo_state_network import ESNRegressor


# ## Load the dataset
# 
# The dataset is already part of scikit-learn and consists of 1797 8x8 images. We are using the dataloader from scikit-learn.

# In[ ]:


digits = load_digits()
data = digits.images
print("Number of digits: {0}".format(len(data)))
print("Shape of digits {0}".format(data[0].shape))


# ## Split dataset in training and test
# 
# We use the OneHotEncoder to transform the target output into one-hot encoded values. 
# 
# Afterwards, we split the dataset into training and test sets. We train the ESN using 50% of the digits and test it using the remaining images. 
# 
# We treat each image as a sequence of 8 feature vectors with 8 dimensions.

# In[ ]:


# Split data into train and test subsets
enc = OneHotEncoder(sparse=False)
y = enc.fit_transform(X=digits.target.reshape(-1, 1))
X_train, X_test, y_train, y_test = train_test_split(data, y, test_size=0.5, shuffle=False)
print("Number of digits in training set: {0}".format(len(X_train)))
print("Shape of digits in training set: {0}".format(X_train[0].shape))
print("Shape of output in training set: {0}".format(y_train[0].shape))
print("Number of digits in test set: {0}".format(len(X_test)))
print("Shape of digits in test set: {0}".format(X_test[0].shape))
print("Shape of output in test set: {0}".format(y_test[0].shape))


# ## Define a user-defined optimization function
# 
# We define a user-defined optimization function to tune the hyper-parameters.

# In[ ]:


def opt_function(base_reg, params, X_train, y_train, X_test, y_test):
    reg = clone(base_reg)
    reg.set_params(**params)
    for X, y in zip(X_train, y_train):
        y = np.repeat(np.atleast_2d(y), repeats=8, axis=0)
        reg.partial_fit(X=X, y=y, update_output_weights=False)
    reg.finalize()
    err_train = []
    for X, y in zip(X_train, y_train):
        y = np.repeat(np.atleast_2d(y), repeats=8, axis=0)
        y_pred = reg.predict(X=X, keep_reservoir_state=False)
        err_train.append(mean_squared_error(y, y_pred))
    err_test = []
    for X, y in zip(X_test, y_test):
        y = np.repeat(np.atleast_2d(y), repeats=8, axis=0)
        y_pred = reg.predict(X=X, keep_reservoir_state=False)
        err_test.append(mean_squared_error(y, y_pred))
    return [np.mean(err_train), np.mean(err_test)]


# ## Set up a basic ESN
# 
# To develop an ESN model for digit recognition, we need to tune several hyper-parameters, e.g., input_scaling, spectral_radius, bias_scaling and leaky integration.
# 
# We follow the way proposed in the introductory paper of PyRCN to optimize hyper-parameters sequentially.
# 
# We start to jointly optimize input_scaling and spectral_radius and therefore deactivate bias connections and leaky integration. This is our base_reg.
# 
# We define the search space for input_scaling and spectral_radius. This is done using best practice and background information from the literature: The spectral radius, the largest absolute eigenvalue of the reservoir matrix, is often smaller than 1. Thus, we can search in a space between 0.0 (e.g. no recurrent connections) and 1.0 (maximum recurrent connections). It is usually recommended to tune the input_scaling factor between 0.1 and 1.0. However, as this is strongly task-dependent, we decided to slightly increase the search space.

# In[ ]:


base_reg = ESNRegressor(k_in = 5, input_scaling = 0.1, spectral_radius = 0.0, bias = 0.0, leakage = 1.0, reservoir_size = 50, 
                   k_res = 5, reservoir_activation = 'tanh', teacher_scaling = 1.0, teacher_shift = 0.0, 
                   bi_directional = False, solver = 'ridge', beta = 5e-3, random_state = 1)

grid = {'input_scaling': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5], 
        'spectral_radius': [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
       }


# ## Optimize input_scaling and spectral_radius
# 
# We use the ParameterGrid from scikit-learn, which converts the grid parameters defined before into a list of dictionaries for each parameter combination. 
# 
# We loop over each entry of the Parameter Grid, set the parameters in reg and fit our model on the training data. Afterwards, we report the MSE on the training and test set.  
# 
#     The lowest training MSE: 0.0725333549527569; parameter combination: {'input_scaling': 0.1, 'spectral_radius': 1.0}
#     The lowest test MSE: 0.0755270784848419; parameter combination: {'input_scaling': 0.1, 'spectral_radius': 0.9}
# 
# We use the best parameter combination from the training set, because we do not want to overfit on the test set.
# 
# As we can see in the python call, we have modified the training procedure: We use "partial_fit" in order to present the ESN all sequences independently from each other. The function "partial_fit" is part of the scikit-learn API. We have added one optional argument "update_output_weights". By default, it is True and thus, after feeding one sequence through the ESN, output weights are computed.
# 
# However, as this is computationally expensive, we can deactivate computing output weights after each sequence by setting "update_output_weights" to False. Now, we simply collect sufficient statistics for the later linear regression. To finish the training process, we call finalize() after passing all sequences through the ESN.

# In[ ]:


t1 = time.time()
losses = Parallel(n_jobs=1, verbose=50)(delayed(opt_function)(base_reg, params, X_train, y_train, X_test, y_test) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest test MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))


# Let us see the impact of parallelization by utilizing all available kernels. One will notice that the optimization runs much faster now.

# In[ ]:


t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_reg, params, X_train, y_train, X_test, y_test) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest test MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))


# ## Update parameter of the basic ESN
# 
# After optimizing input_scaling and spectral_radius, we update our basic ESN with the identified values for input_scaling and spectral_radius. 
# 
# For the next optimization step, we jointly optimize bias and leakage.
# 
# We define the search space for bias and leakage. This is again done using best practice and background information from the literature: The bias often lies in a similar value range as the input scaling. Thus we use exactly the same search space as before. The leakage, the parameter of the leaky integration is defined in (0.0, 1.0]. Thus, we tune the leakage between 0.1 and 1.0.

# In[ ]:


base_reg = ESNRegressor(k_in = 5, input_scaling = 0.1, spectral_radius = 1.0, bias = 0.0, leakage = 1.0, reservoir_size = 50, 
                   k_res = 5, reservoir_activation = 'tanh', teacher_scaling = 1.0, teacher_shift = 0.0, 
                   bi_directional = False, solver = 'ridge', beta = 5e-3, random_state = 1)

grid = {'bias': [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5], 
        'leakage': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
       }


# ## Optimize bias and leakage
# 
# The optimization workflow is exactly the same as before: We define a ParameterGrid, loop over each entry, set the parameters in reg and fit our model on the training data. Afterwards, we report the MSE on the training and test set.  
# 
#     The lowest training MSE: 0.0564864449264251; parameter combination: {'bias': 0.8, 'leakage': 0.2}
#     The lowest test MSE: 0.0626353459066059; parameter combination: {'bias': 0.1, 'leakage': 0.2}
# 
# We use the best parameter combination from the training set, because we do not want to overfit on the test set.
# 
# Note that the bias differs a lot between training and test set. A reason can be that the training set does not completely represent the test set. This should actually be investigated by comparing several train_test_splits, maybe even with other sample sizes.

# In[ ]:


t1 = time.time()
losses = Parallel(n_jobs=1, verbose=50)(delayed(opt_function)(base_reg, params, X_train, y_train, X_test, y_test) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest test MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))


# Let us see the impact of parallelization by utilizing all available kernels. One will notice that the optimization runs much faster now.

# In[ ]:


t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_reg, params, X_train, y_train, X_test, y_test) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest test MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))


# ## Update parameter of the basic ESN
# 
# After optimizing bias and leakage, we update our basic ESN with the identified values for bias and leakage. 
# 
# Finally, we would quickly like to see whether the regularization parameter beta lies in the correct range.
# 
# Typically, it is rather difficult to find a proper search range. Here, we use a very rough logarithmic search space.

# In[ ]:


base_reg = ESNRegressor(k_in = 5, input_scaling = 0.1, spectral_radius = 1.0, bias = 0.8, leakage = 0.1, reservoir_size = 50, 
                   k_res = 5, reservoir_activation = 'tanh', teacher_scaling = 1.0, teacher_shift = 0.0, 
                   bi_directional = False, solver = 'ridge', beta = 5e-3, random_state = 1)

grid = {'beta': [1e-5, 5e-5, 1e-4, 5e-4, 1e-3, 5e-3, 1e-2, 5e-2, 1e-1, 5e-1, 1e0], 
       }


# ## Optimize beta
# 
# The optimization workflow is exactly the same as before: We define a ParameterGrid, loop over each entry, set the parameters in reg and fit our model on the training data. Afterwards, we report the MSE on the training and test set.  
# 
#     The lowest training MSE: 0.055284106204655556; parameter combination: {'beta': 0.0005}
#     The lowest test MSE: 0.06266313201574032; parameter combination: {'beta': 0.001}
# 
# We use the best parameter combination from the test set, because the regularization is responsible to prevent overfitting on the training set. In a running system, of course, we should determine the regularization on a separate validation set.

# In[ ]:


t1 = time.time()
losses = Parallel(n_jobs=1, verbose=50)(delayed(opt_function)(base_reg, params, X_train, y_train, X_test, y_test) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))
print("The lowest training MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest test MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))


# Let us see the impact of parallelization by utilizing all available kernels. One will notice that the optimization runs much faster now.

# In[ ]:


t1 = time.time()
losses = Parallel(n_jobs=-1, verbose=50)(delayed(opt_function)(base_reg, params, X_train, y_train, X_test, y_test) for params in ParameterGrid(grid))
losses = np.asarray(losses)
print("Finished in {0} seconds!".format(time.time() - t1))

print("The lowest training MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 0]), ParameterGrid(grid)[np.argmin(losses[:, 0])]))
print("The lowest test MSE: {0}; parameter combination: {1}".format(np.min(losses[:, 1]), ParameterGrid(grid)[np.argmin(losses[:, 1])]))


# ## Update parameter of the basic ESN
# 
# After optimizing beta, we update our basic ESN with the identified value for beta.
# 
# Note that we have used almost the ideal value already in the beginning. Thus, the impact is rather small.
# 
# Next, we want to measure the classification accuracy. To do that, we compare several reservoir sizes as well as unidirectional and bidirectional architectures.
# 
# Because this is a rather small dataset, we can use rather small reservoir sizes and increase it up to 5000 neurons.

# In[ ]:


base_reg = ESNRegressor(k_in = 5, input_scaling = 0.1, spectral_radius = 1.0, bias = 0.8, leakage = 0.1, reservoir_size = 50, 
                   k_res = 5, reservoir_activation = 'tanh', teacher_scaling = 1.0, teacher_shift = 0.0, 
                   bi_directional = False, solver = 'ridge', beta = 0.0005, random_state = 1)

grid = {'reservoir_size': [50, 100, 200, 400, 500, 800, 1000, 2000, 4000, 5000], 
        'bi_directional': [False, True]
       }


# ## Test the ESN
# 
# In the test case, we use a simple variant of sequence classification:
# 
# The ESN computes the output for each sequence. We integrate the outputs over time and find the highest integrated output index. This is the label of the sequence.
# 
# We store all ground truth labels and the predicted labels for training and test. Then, we use the scikit-learn's classification_report and plot a confusion matrix in order to show the classification performance.
# 
# As can be seen, the reservoir size as a very strong impact on the classification result.

# In[ ]:


for params in ParameterGrid(grid):
    print(params)
    reg = clone(base_reg)
    reg.set_params(**params)
    for X, y in zip(X_train, y_train):
        y = np.repeat(np.atleast_2d(y), repeats=8, axis=0)
        reg.partial_fit(X=X, y=y, update_output_weights=False)
    reg.finalize()
    Y_true_train = []
    Y_pred_train = []
    for X, y in zip(X_train, y_train):
        y_pred = reg.predict(X=X, keep_reservoir_state=False)
        Y_true_train.append(np.argmax(y))
        Y_pred_train.append(np.argmax(y_pred.sum(axis=0)))
    
    Y_true_test = []
    Y_pred_test = []
    for X, y in zip(X_test, y_test):
        y_pred = reg.predict(X=X, keep_reservoir_state=False)
        Y_true_test.append(np.argmax(y))
        Y_pred_test.append(np.argmax(y_pred.sum(axis=0)))
    cm = confusion_matrix(Y_true_train, Y_pred_train)
    cm_display = ConfusionMatrixDisplay(cm, display_labels=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).plot()
    print("Classification training report for estimator %s:\n%s\n"
      % (reg, classification_report(Y_true_train, Y_pred_train)))
    plt.show()
    
    cm = confusion_matrix(Y_true_test, Y_pred_test)
    cm_display = ConfusionMatrixDisplay(cm, display_labels=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).plot()
    print("Classification test report for estimator %s:\n%s\n"
      % (reg, classification_report(Y_true_test, Y_pred_test)))
    plt.show()




import numpy as np
from echo_state_network import ESNRegressor
import matplotlib.pylab as plt
# Latex Backend for Matplotlib
params = {
    'axes.labelsize': 12,
    'font.size': 12,
    'legend.fontsize': 12,
    'xtick.labelsize': 12,
    'ytick.labelsize': 12,
    'font.family': 'serif',
    'text.usetex': True,
    'figure.figsize': [12, 8],
   }

base_esn = ESNRegressor(k_in=10, input_scaling=1.0, spectral_radius=0.9, bias=0.0, ext_bias=False, leakage=0.1,
                        reservoir_size=50, k_res=10, reservoir_activation='tanh', bi_directional=False,
                        teacher_scaling=1.0, teacher_shift=0.0, solver='ridge', beta=1e-6, random_state=10)
X = np.zeros(shape=(100, 10), dtype=int)
X[5] = 1
y = X[:, 0]

base_esn.fit(X=X, y=y, n_jobs=0)
y_pred = base_esn.predict(X=X, keep_reservoir_state=True)

fig, ax = plt.subplots()
im = ax.imshow(base_esn.reservoir_state[:, 1:].T)
ax.set_xlim([0, base_esn.reservoir_state.shape[0]])
ax.set_ylim([0, base_esn.reservoir_state.shape[1] - 1])
ax.set_xlabel(r'k')
ax.set_ylabel(r'X[k]')
plt.grid()
plt.savefig("is_10_sr_09_lr_01.pdf")
exit(0)
